#include "awarp.h"

namespace {
    // Stores the direction from which the computed run comes
    enum class Direction {
        left, top, diagonal
    };
    
    // Cost function between cells for upper bound version of AWarp
    template <typename T>
    inline T ubCosts(const akt::Run<T>& x, int xLength, const akt::Run<T>& y, int yLength, const Direction& dir)
    {
        if (x.val != static_cast<T>(0)) {
            if (y.val != static_cast<T>(0)) {
                return (x.val - y.val) * (x.val - y.val);
            } else {
                if (dir == Direction::left)
                    return x.val * x.val;
                else
                    return x.val * x.val * yLength;
            }
        } else {
            if (y.val != static_cast<T>(0)) {
                if (dir == Direction::top)
                    return y.val * y.val;
                else
                    return y.val * y.val * xLength;
            } else {
                return static_cast<T>(0);
            }
        }
    }
    
    // Cost function between cells for lower bound version of AWarp
    template <typename T>
    inline T lbCosts(const akt::Run<T>& x, int xLength, const akt::Run<T>& y, int yLength, const Direction& dir)
    {
        if (x.val != static_cast<T>(0)) {
            if (y.val != static_cast<T>(0)) {
                return (x.val - y.val) * (x.val - y.val);
            } else {
                if (dir == Direction::top)
                    return x.val * x.val * yLength;
                else
                    return x.val * x.val;
            }
        } else {
            if (y.val != static_cast<T>(0)) {
                if (dir == Direction::left)
                    return y.val * y.val * xLength;
                else
                    return y.val * y.val;
            } else {
                return static_cast<T>(0);
            }
        }
    }
    
    // Initializes the DP array for AWarp
    template <typename T>
    inline std::vector<std::vector<T>> awarpDpInitialize(const std::vector<akt::Run<T>>& x, const std::vector<akt::Run<T>>& y, bool upperBound)
    {
        auto res = std::vector<std::vector<T>>(x.size(), std::vector<T>(y.size()));
        
        res[0][0] = upperBound ? ubCosts(x[0], x[0].boundary + 1, y[0], y[0].boundary + 1, Direction::diagonal) : lbCosts(x[0], x[0].boundary + 1, y[0], y[0].boundary + 1, Direction::diagonal);
        for (int i = 1; i < x.size(); ++i)
            res[i][0] = res[i - 1][0] +
                    (upperBound ? ubCosts(x[i], x[i].boundary - x[i - 1].boundary, y[0], y[0].boundary + 1, Direction::left) : lbCosts(x[i], x[i].boundary - x[i - 1].boundary, y[0], y[0].boundary + 1, Direction::left));
        for (int j = 1; j < y.size(); ++j)
            res[0][j] = res[0][j - 1] +
                    (upperBound ? ubCosts(x[0], x[0].boundary + 1, y[j], y[j].boundary - y[j - 1].boundary, Direction::top) : lbCosts(x[0], x[0].boundary + 1, y[j], y[j].boundary - y[j - 1].boundary, Direction::top));
                    
        return res;
    }
    
} // end unnamed namespace

namespace akt {
    
    // Does simplified RLE where only runs of zeros get compressed
    template <typename T>
    std::vector<Run<T>> toZeroRle(const std::vector<T>& x)
    {
        // Almost a copy-paste of the toRLE function (except for checking whether a potential run is of zeros only)
		std::vector<Run<T>> res;
		for (int i = 0; i < x.size();) {
			int j = i + 1;
            // Check whether we have a potential run of zeros; otherwise we don't compress it at all
            if (x[i] == static_cast<T>(0)) {
                while ((j < x.size()) && (x[j] == x[i]))
                    ++j;
            }
			// Add the run to the result and start the next run
			res.emplace_back(x[i], j - 1);
			i = j;
		}

		return res;
	}

    // Computes the AWarp distance (assumes that Runs of all values different than 0 have length 1)
    // If upperBound is true, returns the upper bound on the true DTW distance; otherwise returns a lower bound
    template <typename T>
    T awarp(const std::vector<Run<T>>& x, const std::vector<Run<T>>& y, bool upperBound)
    {
        auto dp = awarpDpInitialize(x, y, upperBound);
        for (int i = 1; i < dp.size(); ++i) {
            for (int j = 1; j < dp[i].size(); ++j) {
                auto xLen = x[i].boundary - x[i - 1].boundary;
                auto yLen = y[j].boundary - y[j - 1].boundary;
                auto diag = dp[i - 1][j - 1] + 
                    (upperBound ? ubCosts(x[i], xLen, y[j], yLen, Direction::diagonal) : lbCosts(x[i], xLen, y[j], yLen, Direction::diagonal));
                auto top = dp[i][j - 1] +
                    (upperBound ? ubCosts(x[i], xLen, y[j], yLen, Direction::top) : lbCosts(x[i], xLen, y[j], yLen, Direction::top));
                auto left = dp[i - 1][j] +
                    (upperBound ? ubCosts(x[i], xLen, y[j], yLen, Direction::left) : lbCosts(x[i], xLen, y[j], yLen, Direction::left));
                dp[i][j] = std::min(diag, std::min(top, left));
            }
        }
        
        return dp.back().back();
    }

    // The below is so that the code links (a bit of a hack but we only use Run<int> Run<int64_t> and Run<double> anyway
	template std::vector<Run<int>> toZeroRle<int>(const std::vector<int>&);
    template std::vector<Run<int64_t>> toZeroRle<int64_t>(const std::vector<int64_t>&);
    template std::vector<Run<double>> toZeroRle<double>(const std::vector<double>&);
    
    template int awarp<int>(const std::vector<Run<int>>&, const std::vector<Run<int>>&, bool);
    template int64_t awarp<int64_t>(const std::vector<Run<int64_t>>&, const std::vector<Run<int64_t>>&, bool);
    template double awarp<double>(const std::vector<Run<double>>&, const std::vector<Run<double>>&, bool);
    
} // end namespace akt
