#include "compression.h"

namespace {
    // Initializes the two DP arrays for trivial cases (only 1 cluster or at least as many clusters as elements in the (sub-)sequence)
    template <typename T>
    std::pair<std::vector<std::vector<double>>, std::vector<std::vector<std::pair<double, int>>>> compressInit(const std::vector<T>& x, int k)
    {
        if ((x.size() < 1) || (k < 1))
            return std::pair<std::vector<std::vector<double>>, std::vector<std::vector<std::pair<double, int>>>>();
        
        // dpCosts[i][j] stores the optimal cost of (j+1)-clustering of the elements x[0],...,x[i]
        auto dpCosts = std::vector<std::vector<double>>(x.size(), std::vector<double>(k, 0.0));
        // clusterData[i][j] stores a pair of the form [mean in the last cluster, (inclusive) lower bound on the last cluster] for the optimal (j+1)-clustering of the elements x[0],...,x[i]
        auto clusterData = std::vector<std::vector<std::pair<double, int>>>(x.size(), std::vector<std::pair<double, int>>(k));
        // First type of a trivial case: at least as many clusters as elements (in which case the optimal cost is equal to zero)
        for (int i = 0; i < k; ++i) {
            clusterData[i][i] = std::make_pair(x[i], i);
            // Handle the degenerate case: more clusters than elements by having all the superflous clusters be the same as the last useful one
            for (int j = i + 1; j < k; ++j)
                clusterData[i][j] = clusterData[i][j-1];
        }
        // Second type of a trivial case: we only have one cluster
        double mu = x[0]; // Running average
        double ssq = 0.0; // Running sum of squares (of deviations from the mean)
        for (int i = 1; i < x.size(); ++i) {
            ssq += (std::pow(x[i] - mu, 2) / static_cast<double>(i + 1)) * static_cast<double>(i);
            mu = (x[i] + i*mu) / (i + 1);
            dpCosts[i][0] = ssq;
            clusterData[i][0] = std::make_pair(mu, 0);
        }
        
        return std::make_pair(dpCosts, clusterData);
    }

} // end anonymous namespace

namespace akt {
    // Returns all optimal l-clusterings for l in [1, k]
    // The i-th vector in the result containts i elements, each being the length of the corresponding cluster, along with its mean
    // Runs in O(kn^2) time, similar to 1-D k-means DP
    template <typename T>
    std::vector<std::vector<std::pair<int, double>>> compress(const std::vector<T>& x, int k)
    {
        // No need to handle having more clusters than elements
        if (k > x.size())
            k = x.size();
        
        auto [dpCosts, clusterData] = compressInit(x, k);
        
        // i starts at two since all cases with i = 0 or i = 1 are trivial
        for (int i = 2; i < x.size(); ++i) {
            // Handle only the non-trivial cases (more elements than clusters)
            for (int j = 1; j < std::min(i, k); ++j) {
                // Intialize iteration variables: we start with the case where the last cluster consists only of the last element in the prefix subsequence
                dpCosts[i][j] = dpCosts[i - 1][j - 1];
                double mu = static_cast<double>(x[i]); // Running average
                double ssq = 0.0; // Running sum of squares (of the deviations from the cluster mean)
                clusterData[i][j] = std::make_pair(mu, i);
                // Iterate over all possible last clusters and choose the (shortest) optimal one
                for (int l = i - 1; l >= j; --l) {
                    // Compute the new mean, sum of squares and costs for the case where last cluster is [l, i]
                    ssq += (std::pow(x[l] - mu, 2) / static_cast<double>(i - l + 1)) * static_cast<double>(i - l);
                    mu = (x[l] + (i - l)*mu) / (i - l + 1);
                    double cost = dpCosts[l - 1][j - 1] + ssq;
                    if (cost < dpCosts[i][j]) {
                        dpCosts[i][j] = cost;
                        clusterData[i][j] = std::make_pair(mu, l);
                    }
                }
            }
        }
        
        // Reconstruct the optimal solutions for all numbers of clusters [1..k] using the clusterData array
        std::vector<std::vector<std::pair<int, double>>> res;
        res.reserve(k);
        for (int i = 0; i < k; ++i) {
            std::vector<std::pair<int, double>> solutionReversed;
            solutionReversed.reserve(i + 1);
            for(int j = i, upper = x.size() - 1; j >= 0; --j) {
                solutionReversed.emplace_back(upper - clusterData[upper][j].second + 1, clusterData[upper][j].first);
                upper = clusterData[upper][j].second - 1;
            }
            res.emplace_back(solutionReversed.crbegin(), solutionReversed.crend());
        }
        
        return res;
    }

    // Compress x into k runs (or min(k, x.size()), to be precise) using the algorithm implemented in compress()
	template <typename T>
	std::vector<Run<double>> compressToRuns(const std::vector<T>& x, int k)
    {
        k = std::max(1, std::min(static_cast<int>(x.size()), k));

		auto somewhatCompressed = compress(x, k);
		std::vector<Run<double>> res;
        int blockEnd = -1;
		for (auto p : somewhatCompressed[k - 1]) {
			blockEnd += p.first;
            // Note: we use a direct comparison between doubles below, which isn't particularly meaningful
            if ((!res.empty()) && (res.back().val == p.second)) 
                res.back().boundary += p.first;
            else
                res.emplace_back(p.second, blockEnd);
		}

		return res;
    }

    // Needed for the code to link (a bit of a hack but we only use ints, int64_t's and doubles, anyway
	template std::vector<std::vector<std::pair<int, double>>> compress<int>(const std::vector<int>& x, int k);
    template std::vector<std::vector<std::pair<int, double>>> compress<int64_t>(const std::vector<int64_t>& x, int k);
    template std::vector<std::vector<std::pair<int, double>>> compress<double>(const std::vector<double>& x, int k);
	
	template std::vector<Run<double>> compressToRuns<int>(const std::vector<int>& x, int k);
	template std::vector<Run<double>> compressToRuns<int64_t>(const std::vector<int64_t>& x, int k);
	template std::vector<Run<double>> compressToRuns<double>(const std::vector<double>& x, int k);

} // end namespace akt
