#include "dtwAlgorithms.h"

#include <iostream>
#include <numeric>

namespace {
    using akt::COST_INFINITE;
    using akt::IntersectionPoint;
    using akt::Intersections;
    using akt::Position;
    using akt::Run;
    
    // Returns to the cost if path comes from the left
    // Infinity if path can't come from the left
    template <typename T>
    inline T leftCost(const std::vector<Run<T>>& x, const std::vector<Run<T>>& y, const Intersections<T>& dp, const IntersectionPoint<T>& ip, Position curBoundaries, T curBlockCost)
    {
        // Check if ip lies on an x-boundary (as the left cost is only relevant for those), and whether the cost to the left exists and is finite
        if ((ip.loc.x != x[curBoundaries.x].boundary) || (ip.posInBoundary.x <= 0)
            || (dp.xIntersections[curBoundaries.x][ip.posInBoundary.x - 1].cost == COST_INFINITE<T>))
            return COST_INFINITE<T>;
        // The path could come from the left; return its cost
        return dp.xIntersections[curBoundaries.x][ip.posInBoundary.x - 1].cost
            + (ip.loc.y - dp.xIntersections[curBoundaries.x][ip.posInBoundary.x - 1].loc.y)
                * curBlockCost;
    }

    // Returns to the cost if path comes from below
    // Infinity if path can't come from the below
    template <typename T>
    inline T belowCost(const std::vector<Run<T>>& x, const std::vector<Run<T>>& y, const Intersections<T>& dp, const IntersectionPoint<T>& ip, Position curBoundaries, T curBlockCost)
    {
        // Check if ip lies on on a y-boundary (as the below cost is only relevant for those)
        if (ip.loc.y != y[curBoundaries.y].boundary)
            return COST_INFINITE<T>;
        
        // Compute the index of the nearest x-intersection below ip
        // (it may stay as -1 if we can be certain that it is irrelevant to the cost)
        auto belowIdX = -1;
        if ((ip.prevBelowIdx.x >= 0) && (curBoundaries.x >= 1)
            && (ip.prevBelowIdx.x < dp.xIntersections[curBoundaries.x - 1].size())) {
            if (dp.xIntersections[curBoundaries.x - 1][ip.prevBelowIdx.x].loc.y 
                == ip.loc.y)
                belowIdX = ip.prevBelowIdx.x;
            else if (((ip.prevBelowIdx.x + 1) < dp.xIntersections[curBoundaries.x - 1].size()) 
                && (dp.xIntersections[curBoundaries.x - 1][ip.prevBelowIdx.x + 1].loc.y 
                == ip.loc.y))
                belowIdX = ip.prevBelowIdx.x + 1;
        }
        
        // Compute the index of the nearest y-intersection below ip
        // (it may stay as -1 if we can be certain that it is irrelevant to the cost)
        auto belowIdY = -1;
        if ((ip.prevBelowIdx.y >= 0) 
            && (ip.prevBelowIdx.y < dp.yIntersections[curBoundaries.y].size()))
            belowIdY = ip.prevBelowIdx.y;
        
        if (belowIdX < 0) {
            if ((belowIdY < 0) 
                || (dp.yIntersections[curBoundaries.y][belowIdY].cost == COST_INFINITE<T>))
                return COST_INFINITE<T>;
            return dp.yIntersections[curBoundaries.y][belowIdY].cost
                + (ip.loc.x - dp.yIntersections[curBoundaries.y][belowIdY].loc.x)
                * curBlockCost;
        }
        if ((belowIdY < 0) 
            || (dp.yIntersections[curBoundaries.y][belowIdY].loc.x <= dp.xIntersections[curBoundaries.x - 1][belowIdX].loc.x)) {
            if (dp.xIntersections[curBoundaries.x - 1][belowIdX].cost == COST_INFINITE<T>)
                return COST_INFINITE<T>;
            return dp.xIntersections[curBoundaries.x - 1][belowIdX].cost
                + (ip.loc.x - dp.xIntersections[curBoundaries.x - 1][belowIdX].loc.x)
                * curBlockCost;
        }
        
        return dp.yIntersections[curBoundaries.y][belowIdY].cost
                + (ip.loc.x - dp.yIntersections[curBoundaries.y][belowIdY].loc.x)
                * curBlockCost;
    }

    // Returns to the cost if path comes from the diagonal direction
    // Infinity if path can't come from the diagonal direction
    template <typename T>
    inline T diagonalCost(const std::vector<Run<T>>& x, const std::vector<Run<T>>& y, const Intersections<T>& dp, const IntersectionPoint<T>& ip, Position curBoundaries, T curBlockCost)
    {
        if (ip.prevPosInBoundary.x < 0) {
            if ((ip.prevPosInBoundary.y < 0)
                || (dp.yIntersections[curBoundaries.y - 1][ip.prevPosInBoundary.y].cost == COST_INFINITE<T>))
                return COST_INFINITE<T>;
            return dp.yIntersections[curBoundaries.y - 1][ip.prevPosInBoundary.y].cost
                + (ip.loc.y - dp.yIntersections[curBoundaries.y - 1][ip.prevPosInBoundary.y].loc.y)
                * curBlockCost;
        }
        
        if (dp.xIntersections[curBoundaries.x - 1][ip.prevPosInBoundary.x].cost == COST_INFINITE<T>)
            return COST_INFINITE<T>;
        return dp.xIntersections[curBoundaries.x - 1][ip.prevPosInBoundary.x].cost
            + (ip.loc.x - dp.xIntersections[curBoundaries.x - 1][ip.prevPosInBoundary.x].loc.x)
                * curBlockCost;
    }

    // Handles the trivial case when one sequence is very degenerate (i.e. has exactly one element)
    template <typename T>
    T handleTrivial(const std::vector<Run<T>>& x, const std::vector<Run<T>>& y)
    {
        if (x.size() == 1) {
            T res = (y[0].boundary + 1) * (x.front().val - y[0].val) * (x.front().val - y[0].val);
            for (int i = 1; i < y.size(); ++i)
                res += (y[i].boundary - y[i - 1].boundary) * (x.front().val - y[i].val) * (x.front().val - y[i].val);
            return res;
        } else {
            T res = (x[0].boundary + 1) * (y.front().val - x[0].val) * (y.front().val - x[0].val);
            for (int i = 1; i < x.size(); ++i)
                res += (x[i].boundary - x[i - 1].boundary) * (y.front().val - x[i].val) * (y.front().val - x[i].val);
            return res;
        }
    }
    
    // Computes a lower bound on the DTW distance with BDTW
    template <typename T>
    T dtwDistanceApp_blockDTW_min(const std::vector<Run<T>>& x, const std::vector<Run<T>>& y)
    {
        auto dp = std::vector<std::vector<T>>(x.size(), std::vector<T>(y.size()));
        dp[0][0] = (std::min(x[0].boundary, y[0].boundary) + 1) 
                    * (x[0].val - y[0].val) * (x[0].val - y[0].val);
        for (int i = 1; i < x.size(); ++i) {
            dp[i][0] = dp[i - 1][0] + (x[i].boundary - x[i - 1].boundary) 
                        * (x[i].val - y[0].val) * (x[i].val - y[0].val);
        }
        for (int j = 1; j < y.size(); ++j) {
            dp[0][j] = dp[0][j - 1] + (y[j].boundary - y[j - 1].boundary)
                        * (x[0].val - y[j].val) * (x[0].val - y[j].val);
        }
        for (int i = 1; i < x.size(); ++i) {
            for (int j = 1; j < y.size(); ++j) {
                auto len = std::min(x[i].boundary - x[i - 1].boundary, 
                                    y[j].boundary - y[j - 1].boundary);
                auto blockValue = (x[i].val - y[j].val) * (x[i].val - y[j].val);
                dp[i][j] = std::min(dp[i-1][j] + (x[i].boundary - x[i - 1].boundary) * blockValue, dp[i][j - 1] + (y[j].boundary - y[j - 1].boundary) * blockValue);
                dp[i][j] = std::min(dp[i][j], dp[i - 1][j - 1] + len * blockValue);
            }
        }
        return dp.back().back();
    }
    
    // Computes an upper bound on the DTW distance with BDTW
    template <typename T>
    T dtwDistanceApp_blockDTW_max(const std::vector<Run<T>>& x, const std::vector<Run<T>>& y)
    {
        auto dp = std::vector<std::vector<T>>(x.size(), std::vector<T>(y.size()));
        dp[0][0] = (std::max(x[0].boundary, y[0].boundary) + 1) 
                    * (x[0].val - y[0].val) * (x[0].val - y[0].val);
        for (int i = 1; i < x.size(); ++i) {
            dp[i][0] = dp[i - 1][0] + (x[i].boundary - x[i - 1].boundary) 
                        * (x[i].val - y[0].val) * (x[i].val - y[0].val);
        }
        for (int j = 1; j < y.size(); ++j) {
            dp[0][j] = dp[0][j - 1] + (y[j].boundary - y[j - 1].boundary)
                        * (x[0].val - y[j].val) * (x[0].val - y[j].val);
        }
        for (int i = 1; i < x.size(); ++i) {
            for (int j = 1; j < y.size(); ++j) {
                auto len = std::max(x[i].boundary - x[i - 1].boundary, 
                                    y[j].boundary - y[j - 1].boundary);
                auto blockValue = (x[i].val - y[j].val) * (x[i].val - y[j].val);
                dp[i][j] = std::min(dp[i-1][j] + (x[i].boundary - x[i - 1].boundary) * blockValue, dp[i][j - 1] + (y[j].boundary - y[j - 1].boundary) * blockValue);
                dp[i][j] = std::min(dp[i][j], dp[i - 1][j - 1] + len * blockValue);
            }
        }
        return dp.back().back();
    }
    
    // Precomputes the block costs
    template <typename T>
    inline std::vector<std::vector<T>> precomputeBlockCosts(const std::vector<Run<T>>& x, const std::vector<Run<T>>& y)
    {
        std::vector<std::vector<T>> res;
        res.reserve(x.size());
        for (const auto& xr : x) {
            std::vector<T> row;
            row.reserve(y.size());
            for (const auto& yr : y)
                row.push_back((xr.val - yr.val) * (xr.val - yr.val));
            res.push_back(std::move(row));
        }
        
        return res;
    }
    
} // end anonymous namespace

namespace akt {

	// Computes the square of the DTW-distance between x and y using the standard DP algorithm
    template <typename T>
	T dtwDistance_basicDP(const std::vector<T>& x, const std::vector<T>& y)
	{
		if ((x.size() < 1) || (y.size() < 1))
			return COST_INFINITE<T>;

		auto dp = std::vector<std::vector<T>>(x.size(), std::vector<T>(y.size(), -1));
		// Initialize the boundary values
		dp[0][0] = (x[0] - y[0]) * (x[0] - y[0]);
		for (int i = 1; i < x.size(); ++i)
			dp[i][0] = dp[i - 1][0] + ((x[i] - y[0]) * (x[i] - y[0]));
		for (int j = 1; j < y.size(); ++j)
			dp[0][j] = dp[0][j - 1] + ((x[0] - y[j]) * (x[0] - y[j]));
		// Fill the table
		for (int i = 1; i < x.size(); ++i) {
			for (int j = 1; j < y.size(); ++j) {
				// Initial cost of matching i and j
				dp[i][j] = (x[i] - y[j]) * (x[i] - y[j]);
				// Add the minimal cost of the sequence before
				dp[i][j] += std::min(dp[i - 1][j - 1], std::min(dp[i - 1][j], dp[i][j - 1]));
			}
		}

		return dp.back().back();
	}

	// Uses the O(kl*log(kl) + \kappa) algorithm to compute the DTW distance
	template <typename T>
	T dtwDistance_RLEDTW(const std::vector<Run<T>>& x, const std::vector<Run<T>>& y)
	{
		if ((x.size() <= 0) || (y.size() <= 0))
			return COST_INFINITE<T>;
		if (((x.size() == 1) && (x.front().boundary == 0)) || ((y.size() == 1) && (y.front().boundary == 0)))
			return handleTrivial(x, y);

		auto dp = getIntersectionPoints(x, y);

		return dtwDistance_RLEDTW(x, y, std::move(dp));
	}
	
	// Computes the dtwDistance using the RLEDTW algorithm given the intersection points
	template <typename T>
	T dtwDistance_RLEDTW(const std::vector<Run<T>>& x, const std::vector<Run<T>>& y, Intersections<T> dp)
	{
		if ((x.size() <= 0) || (y.size() <= 0))
			return COST_INFINITE<T>;
		if (((x.size() == 1) && (x.front().boundary == 0)) || ((y.size() == 1) && (y.front().boundary == 0)))
			return handleTrivial(x, y);

        //auto blockCosts = precomputeBlockCosts(x, y);
		std::vector<int> yPoses;
		yPoses.reserve(y.size());
		for (const auto& ys : dp.yIntersections)
			yPoses.push_back(static_cast<int>(ys.size()) - 1);
		// Go through the cells in a (-n admittedly weird) topological order
		for (int xIdx = 0; xIdx < x.size(); ++xIdx) {
			int curBoundary = x[xIdx].boundary;
			for (int yIdx = 0; yIdx < y.size(); ++yIdx) {
				while ((yPoses[yIdx] >= 0)
					&& (dp.yIntersections[yIdx][yPoses[yIdx]].loc.x < curBoundary)) {
					// Handle the point
					auto cost = std::min(diagonalCost(x, y, dp, dp.yIntersections[yIdx][yPoses[yIdx]], Position{ xIdx, yIdx }, (x[xIdx].val - y[yIdx].val) * (x[xIdx].val - y[yIdx].val)), belowCost(x, y, dp, dp.yIntersections[yIdx][yPoses[yIdx]], Position{ xIdx, yIdx }, (x[xIdx].val - y[yIdx].val) * (x[xIdx].val - y[yIdx].val)));
					// So that start initialization works
					if (dp.yIntersections[yIdx][yPoses[yIdx]].cost < 0)
						dp.yIntersections[yIdx][yPoses[yIdx]].cost = cost;
                    
					--(yPoses[yIdx]);
				}
			}
			int yIdx = 0;
			for (auto& ip : dp.xIntersections[xIdx]) {
				while (y[yIdx].boundary < ip.loc.y)
					++yIdx;
				// Handle the point
				auto cost = std::min(diagonalCost(x, y, dp, ip, Position{ xIdx, yIdx }, (x[xIdx].val - y[yIdx].val) * (x[xIdx].val - y[yIdx].val)), leftCost(x, y, dp, ip, Position{ xIdx, yIdx }, (x[xIdx].val - y[yIdx].val) * (x[xIdx].val - y[yIdx].val)));
				cost = std::min(cost, belowCost(x, y, dp, ip, Position{ xIdx, yIdx }, (x[xIdx].val - y[yIdx].val) * (x[xIdx].val - y[yIdx].val)));
                // So that start initialization works
				if (ip.cost < 0)
					ip.cost = cost;
			}
		}

		return dp.xIntersections.back().back().cost;
	}

	// Implements the BlockDTW algorithm (which considers only the corners of boundaries in its calculations
	// If max is true, it provides an upper bound on the DTW distance; otherwise it provides a lower bound on the DTW distance
	template <typename T>
	T dtwDistanceApp_blockDTW(const std::vector<Run<T>>& x, const std::vector<Run<T>>& y, bool max)
	{
		if ((x.size() < 1) || (y.size() < 1))
			return COST_INFINITE<T>;

		if (max)
			return dtwDistanceApp_blockDTW_max(x, y);
		return dtwDistanceApp_blockDTW_min(x, y);
	}

	// Needed for the code to link (a bit of a hack but we only use ints, int64_t's and doubles, anyway
	template int dtwDistance_basicDP<int>(const std::vector<int>&, const std::vector<int>&);
    template int64_t dtwDistance_basicDP<int64_t>(const std::vector<int64_t>&, const std::vector<int64_t>&);
    template double dtwDistance_basicDP<double>(const std::vector<double>&, const std::vector<double>&);
    
    template int dtwDistance_RLEDTW<int>(const std::vector<Run<int>>&, const std::vector<Run<int>>&);
    template int64_t dtwDistance_RLEDTW<int64_t>(const std::vector<Run<int64_t>>&, const std::vector<Run<int64_t>>&);
    template double dtwDistance_RLEDTW<double>(const std::vector<Run<double>>&, const std::vector<Run<double>>&);
    
    template int dtwDistance_RLEDTW<int>(const std::vector<Run<int>>&, const std::vector<Run<int>>&, Intersections<int>);
    template int64_t dtwDistance_RLEDTW<int64_t>(const std::vector<Run<int64_t>>&, const std::vector<Run<int64_t>>&, Intersections<int64_t>);
    template double dtwDistance_RLEDTW<double>(const std::vector<Run<double>>&, const std::vector<Run<double>>&, Intersections<double>);
    
    template int dtwDistanceApp_blockDTW<int>(const std::vector<Run<int>>&, const std::vector<Run<int>>&, bool);
    template int64_t dtwDistanceApp_blockDTW<int64_t>(const std::vector<Run<int64_t>>&, const std::vector<Run<int64_t>>&, bool);
    template double dtwDistanceApp_blockDTW<double>(const std::vector<Run<double>>&, const std::vector<Run<double>>&, bool);
    
}	// end namespace akt
