#include <algorithm>
#include <chrono>
#include <cmath>
#include <fstream>
#include <iostream>
#include <limits>
#include <random>
#include <string>
#include <sstream>
#include <utility>
#include <vector>

#include <boost/program_options.hpp>

#include "awarp.h"
#include "compression.h"
#include "dtwAlgorithms.h"
#include "sparsedtw.h"
#include "utility.h"

#define STATISTICS 1
#include "../../lin_kapp_impl/dtw.hpp"
#include "../../lin_kapp_impl/utils.hpp"

namespace po = boost::program_options;

namespace {
	struct BenchmarkSettings {
		bool runBasic = false;
		bool runBlock = true;
		bool runAwarp = false;
		bool runSparse = false;
        bool runKappa = true;
		double compressionPercentage = -1.0;
        // The maximum number of sequences to compare--if more sequences than that are input, they won't be run benchmarks on. -1 means "compare all"
        int maxSequences = -1;
        // The number of elements to ignore at the start of each sequence read (in UCR format, it's usually 1, because they start with class labels)
        int elementsInSequenceToIgnore = 1;
	};
    
    template <typename T>
	struct BenchmarkResults {
		// Times are all in seconds
		double basicTime = -1.0; // O(mn) alg
		double blockTime = -1.0; // BTDW time (for one of the two runs, but since they are same computationally, it doesn't really matter which)
		double intersectionPointsTime = -1.0; // Time taken for computing intersection points
		double awarpTime = -1.0; // Total time for AWarp (for one of the two runs)
		double sparseTime = -1.0; // Total time for SparseDTW
		double kappaTime = -1.0; // Total time for O(kappa) algorithm
		T dtwDist = -1; // The exact DTW distance
		T bdtw_lower = -1; // The lower bound computed by BTDW
		T bdtw_upper = -1; // The upper bound computed by BTDW
		T awarp_lower = -1; // The lower bound computed by AWarp
		T awarp_upper = -1; // The upper bound computed by AWarp
		int64_t kappa = -1; // The total number of intersection points
	};

    template <typename T>
	std::ostream& operator<<(std::ostream& os, const BenchmarkResults<T>& br)
	{
		os << br.dtwDist << ' ' << br.bdtw_lower << ' ' << br.bdtw_upper << ' ' << br.awarp_lower << ' ' << br.awarp_upper << ' ' << br.kappa << ' ';
		os << br.basicTime << ' ' << br.blockTime << ' ' << br.awarpTime << ' ' << br.sparseTime << ' ' << br.intersectionPointsTime << ' ' << br.kappaTime;

		return os;
	}
    
    template <typename T>
	BenchmarkResults<T> benchmark(const std::vector<akt::Run<T>>& x, const std::vector<akt::Run<T>>& y, const BenchmarkSettings& bs)
	{
		BenchmarkResults<T> res;

		auto xFlat = fromRLE(x);
		auto yFlat = fromRLE(y);
		
        if (bs.runBasic) {
			auto start = std::chrono::high_resolution_clock::now();
			res.dtwDist = akt::dtwDistance_basicDP(xFlat, yFlat);
			auto end = std::chrono::high_resolution_clock::now();
			std::chrono::duration<double> time = end - start;
			res.basicTime = time.count();
		}

		/*if (bs.runRle) {
            auto start = std::chrono::high_resolution_clock::now();
			auto ips = akt::getIntersectionPoints(x, y);
			auto end = std::chrono::high_resolution_clock::now();
			std::chrono::duration<double> time = end - start;
			res.intersectionPointsTime = time.count();
            start = std::chrono::high_resolution_clock::now();
			auto rleDist = akt::dtwDistance_RLEDTW(x, y, ips);
			end = std::chrono::high_resolution_clock::now();
			time = end - start;
			res.rleTime = res.intersectionPointsTime + time.count();
			//res.kappa = akt::kappa(ips);
			//if (bs.runBasic && (rleDist != res.dtwDist))
			//	std::cerr << "DTW distance discrepancy: \nbasic: " << res.dtwDist << '\n' << "rle: " << rleDist << '\n';
            res.dtwDist = rleDist;
		}*/
		
        if (bs.runKappa) {
            auto kappaX = time_seq_to_RLE(xFlat);
            auto kappaY = time_seq_to_RLE(yFlat);
            auto instance = kappa::Solver<T>(kappaX, kappaY);
            auto start = std::chrono::high_resolution_clock::now();
			auto kappaDist = instance.compute_dtw();
			auto end = std::chrono::high_resolution_clock::now();
			std::chrono::duration<double> time = end - start;
			res.kappaTime = time.count();
            res.kappa = instance.get_kappa();
            //if (bs.runBasic && (kappaDist != res.dtwDist))
            //  std::cerr << "KappaDTW discrepancy:\nbasic: " << res.dtwDist << "\nKappaDTW:" << kappaDist << '\n';
            res.dtwDist = kappaDist;
        }

		if (bs.runBlock) {
			auto start = std::chrono::high_resolution_clock::now();
			res.bdtw_lower = akt::dtwDistanceApp_blockDTW(x, y, false);
			auto end = std::chrono::high_resolution_clock::now();
			std::chrono::duration<double> time = end - start;
			res.blockTime = time.count();
			//if (bs.runBasic && (res.bdtw_lower > res.dtwDist))
			//	std::cerr << "BTDW discrepancy:\ndistance is: " << res.dtwDist << ", but BDTW outputs " << res.bdtw_lower << " as a lower bound\n";

			res.bdtw_upper = akt::dtwDistanceApp_blockDTW(x, y, true);
			//if (bs.runBasic && (res.bdtw_upper < res.dtwDist))
			//	std::cerr << "BTDW discrepancy:\ndistance is: " << res.dtwDist << ", but BDTW outputs " << res.bdtw_upper << " as an uppper bound\n";
		}
        
		if (bs.runAwarp) {
            auto awarpX = akt::toZeroRle(xFlat);
            auto awarpY = akt::toZeroRle(yFlat);
			auto start = std::chrono::high_resolution_clock::now();
			res.awarp_lower = akt::awarp(awarpX, awarpY, false);
			auto end = std::chrono::high_resolution_clock::now();
			std::chrono::duration<double> time = end - start;
			res.awarpTime = time.count();
			//if (bs.runBasic && (res.awarp_lower > res.dtwDist))
			//	std::cerr << "AWarp discrepancy:\ndistance is: " << res.dtwDist << ", but AWarp outputs " << res.awarp_lower << " as a lower bound\n";

			res.awarp_upper = akt::awarp(awarpX, awarpY, true);
			//if (bs.runBasic && (res.awarp_upper < res.dtwDist))
			//	std::cerr << "AWarp discrepancy:\ndistance is: " << res.dtwDist << ", but AWarp outputs " << res.awarp_upper << " as an uppper bound\n";
		}

		if (bs.runSparse) {
			auto start = std::chrono::high_resolution_clock::now();
			auto sparseDist = akt::sparseDtw(akt::toSparseObservation(xFlat), akt::toSparseObservation(yFlat));
			auto end = std::chrono::high_resolution_clock::now();
			std::chrono::duration<double> time = end - start;
			res.sparseTime = time.count();
			//if (bs.runBasic && (sparseDist != res.dtwDist))
			//	std::cerr << "SparseDTW discrepancy:\nbasic: " << res.dtwDist << "\nSparseDTW:" << sparseDist << '\n';
		}

		return res;
	}
    
    // Reads in the runs from the given input stream
    template <typename T>
    std::vector<std::vector<akt::Run<T>>> read(std::istream& is, int maxSequencesToRead)
    {
        std::vector<std::vector<akt::Run<T>>> res;
        if (maxSequencesToRead == 0)
            return res;
        
        int sequencesRead = 0;
        for (auto opt = akt::readRuns<T>(is); opt && ((maxSequencesToRead < 0) || (sequencesRead < maxSequencesToRead)); opt = akt::readRuns<T>(is)) {
            if ((opt.value().size() > 0) && (opt.value().back().boundary >= 0)) {
                res.push_back(opt.value());
                ++sequencesRead;
            }
        }
        return res;
    }
    // Reads in the runs from the given input stream and "flattens" them
    template <typename T>
    std::vector<std::vector<T>> readFlat(std::istream& is, int maxSequencesToRead)
    {
        std::vector<std::vector<T>> res;
        if (maxSequencesToRead == 0)
            return res;
        
        int sequencesRead = 0;
        for (auto opt = akt::readRuns<T>(is); opt && ((maxSequencesToRead < 0) || (sequencesRead < maxSequencesToRead)); opt = akt::readRuns<T>(is)) {
            if ((opt.value().size() > 0) && (opt.value().back().boundary >= 0)) {
                res.push_back(akt::fromRLE(opt.value()));
                ++sequencesRead;
            }
        }
        return res;
    }
    
    // Reads in the newline-separated non-empty sequences of numbers and converts them to RLE
    // Ignores first toIgnoreInSequence elements in each sequence read
    template <typename T>
    std::vector<std::vector<akt::Run<T>>> readSequences(std::istream& is, int maxSequencesToRead, int toIgnoreInSequence)
    {
        std::vector<std::vector<akt::Run<T>>> res;
        if (maxSequencesToRead == 0)
            return res;
        
        int sequencesRead = 0;
        for (std::string line; std::getline(is, line) && ((maxSequencesToRead < 0) || (sequencesRead < maxSequencesToRead)); ) {
            auto iss = std::istringstream(line);
            std::vector<T> sequence;
            std::copy(std::istream_iterator<T>(iss), std::istream_iterator<T>(),
              std::back_inserter(sequence));
            if (sequence.size() > toIgnoreInSequence) {
                sequence = std::vector<T>(sequence.cbegin() + toIgnoreInSequence, sequence.cend());
                res.push_back(akt::toRLE(sequence));
                ++sequencesRead;
            }
        }
        
        return res;
    }
    
    // Reads in the newline-separated non-empty sequences of numbers (ignores the first one)
    // Ignores first toIgnoreInSequence elements in each sequence read
    // Note: this function does the same thing as readSequences() except for converting to RLE
    template <typename T>
    std::vector<std::vector<T>> readFlatSequences(std::istream& is, int maxSequencesToRead, int toIgnoreInSequence)
    {
        std::vector<std::vector<T>> res;
        if (maxSequencesToRead == 0)
            return res;
        
        int sequencesRead = 0;
        for (std::string line; std::getline(is, line) && ((maxSequencesToRead < 0) || (sequencesRead < maxSequencesToRead)); ) {
            auto iss = std::istringstream(line);
            std::vector<T> sequence;
            std::copy(std::istream_iterator<T>(iss), std::istream_iterator<T>(),
              std::back_inserter(sequence));
            if (sequence.size() > toIgnoreInSequence) {
                res.emplace_back(sequence.cbegin() + toIgnoreInSequence, sequence.cend());
                ++sequencesRead;
            }
        }
        
        return res;
    }
    
    template <typename T>
    void runBenchmarks(std::vector<std::vector<akt::Run<T>>>& runs, const BenchmarkSettings& bs) 
    {
        // Dirty hack to make SparseDTW work correctly
        for (auto& run : runs) {
            while ((!run.empty()) && (run.back().val == 0))
                run.pop_back();
        }
        // Ignore empty runs
        runs.erase(std::remove_if(runs.begin(), runs.end(),
                                  [](const auto& r) {return r.size() == 0;}), runs.end());
        
        for (int i = 0; i < runs.size(); ++i) {
            for (int j = i + 1; j < runs.size(); ++j) {
                auto benchRes = benchmark(runs[i], runs[j], bs);
                std::cout << i + 1 << "-" << j + 1 << ' ' << benchRes << '\n';
            }
        }
    }
    
    // // Reads in the sequences of the given type, compresses the sequence runs the benchmarks, outputs its result for each pair of signals
    // If filename == "" then reads from stdin
    void readCompressBenchmark(const std::string& filename, bool ucr, const BenchmarkSettings& bs)
    {
        std::vector<std::vector<double>> flats;
        
        if (filename == "") {
            std::clog << "Reading sequences from stdin... ";
            if (ucr)
                flats = readFlatSequences<double>(std::cin, bs.maxSequences, bs.elementsInSequenceToIgnore);
            else 
                flats = readFlat<double>(std::cin, bs.maxSequences);
        } else {
            std::ifstream ifs {filename};
            if (!ifs) {
                std::cerr << "Error opening " << filename << '\n';
                exit(1);
            }
            std::clog << "Reading sequences from " << filename << "... ";
            if (ucr)
                flats = readFlatSequences<double>(ifs, bs.maxSequences, bs.elementsInSequenceToIgnore);
            else
                flats = readFlat<double>(ifs, bs.maxSequences);
        }
        
        std::vector<std::vector<akt::Run<double>>> sequences;
        for (const auto& flat : flats) {
            int noRuns = std::max(1, static_cast<int>(std::round(bs.compressionPercentage * flat.size())));
            auto runs = akt::compressToRuns(flat, noRuns);
            sequences.push_back(std::move(runs));
        }
        std::clog << "Done." << std::endl;
        
        runBenchmarks(sequences, bs);
    }
    
    // Reads in the sequences of the given type, runs the benchmarks, outputs its result for each pair of signals
    // If filename == "" then reads from stdin
    template <typename T>
    void readAndBenchmark(const std::string& filename, bool ucr, const BenchmarkSettings& bs)
    {
        std::vector<std::vector<akt::Run<T>>> runs;
        
        if (filename == "") {
            std::clog << "Reading sequences from stdin... ";
            if (ucr)
                runs = readSequences<T>(std::cin, bs.maxSequences, bs.elementsInSequenceToIgnore);
            else 
                runs = read<T>(std::cin, bs.maxSequences);
        } else {
            std::ifstream ifs {filename};
            if (!ifs) {
                std::cerr << "Error opening " << filename << '\n';
                exit(1);
            }
            std::clog << "Reading sequences from " << filename << "... ";
            if (ucr)
                runs = readSequences<T>(ifs, bs.maxSequences, bs.elementsInSequenceToIgnore);
            else
                runs = read<T>(ifs, bs.maxSequences);
        }
        std::clog << "Done." << std::endl;
        
        runBenchmarks(runs, bs);
    }
}

int main(int argc, char** argv)
{
    std::string filename;
	BenchmarkSettings bs;
    po::options_description desc("usage: dtwBenchmark [options]\nRuns the different DTW algorithms on all possible pairs of the newline-separated non-empty sequences input. If no file name is provided, reads those from stdin. Default input format is that used by dtwGenerator. For UCR format, use -u (and possibly -t). Available options");
	desc.add_options()
		("help,h", "write help message")
		("filename,f", po::value<std::string>(&filename), "filename from which sequences should be read; can also be given as a positional parameter")
		("integer,i", "read sequence as 64-bit integers (by default sequences are assumed to consist of doubles)")
		("ucr,u", "read the data in the UCR format (sequences of floating-point values separated by newlines)")
        ("not-truncate-first,t", "if -u is used, the first element of each sequence is ignored by default (class label). if -t is used, this first element will instead be also read")
		("compress,c", po::value<double>(&(bs.compressionPercentage)), "compress the sequences to (approximately) the given percentage (so -c 0.4 reduces the size of the inputs by 60 %), input should be in [0, 1]")
		("run-naive,n", po::value<bool>(&(bs.runBasic))->default_value(false), "run the standard O(n^2) DP algorithm")
		("run-bdtw,b", po::value<bool>(&(bs.runBlock))->default_value(true), "run the BDTW approximation algorithm")
		("run-awarp,a", po::value<bool>(&(bs.runAwarp))->default_value(false), "run the AWarp algorithm")
		("run-sparse,s", po::value<bool>(&(bs.runSparse))->default_value(false), "run the SparseDTW algorithm")
        ("run-kappa,k", po::value<bool>(&(bs.runKappa))->default_value(true), "run the O(kappa) algorithm")
        ("sequences-to-compare,l", po::value<int>(&(bs.maxSequences))->default_value(-1), "the maximum number of sequences to be run comparisons on. if more are input, they won't be run benchmarks on. if set to a negative value, all sequences will be run benchmarks on");
    
    po::positional_options_description posDesc;
    posDesc.add("filename", 1);
    
    po::variables_map vm;
	po::store(po::command_line_parser(argc, argv).options(desc).positional(posDesc).run(), vm);
	po::notify(vm);
    
    if (vm.count("help")) {
        std::cout << desc << '\n';
        std::cout << "Note: runs of zeros at the end of a sequence are always ignored (and hence so are sequences consisting of zeros only)\n";
        std::cout << "Output format:\n";
        std::cout << "pair-of-sequences dtw-distance bdtw-lower-bound bdtw-upper-bound awarp-lower-bound awarp-upper-bound kappa time-for-basic-dp time-for-bdtw time-for-awarp time-for-sparsedtw time-for-intersection-points time-for-kappa-alg\n";
        return 0;
    }
    
    bs.elementsInSequenceToIgnore = (vm.count("not-truncate-first") <= 0);
    bool ucr = vm.count("ucr") > 0;
	if (bs.compressionPercentage < 0)
		bs.compressionPercentage = -1.0;
	bs.compressionPercentage = std::min(1.0, bs.compressionPercentage);
    if (bs.compressionPercentage >= 0.0)
        readCompressBenchmark(filename, ucr, bs);
    else if (vm.count("integer"))
        readAndBenchmark<int64_t>(filename, ucr, bs);
    else
        readAndBenchmark<double>(filename, ucr, bs);
}
