#include <algorithm>
#include <iostream>
#include <limits>
#include <random>
#include <utility>
#include <vector>

#include <boost/program_options.hpp>

#include "utility.h"

namespace po = boost::program_options;

using akt::RunI;

namespace akt {
	template <class ValueDist, class RunLenDist>
	std::vector<std::vector<RunI>> generateSequences(ValueDist vd, RunLenDist rd, int seqLen, int noSeqs)
	{
		std::vector<std::vector<RunI>> res;
		res.reserve(noSeqs);
		for (int i = 0; i < noSeqs; ++i)
			res.push_back(randomSequence(vd, rd, seqLen));
		return res;
	}

} // end namespace akt

int main(int argc, char** argv)
{
    int n = -1, len = -1;
    std::vector<int> distributionParameters;
    std::vector<int> valRange;
    po::options_description desc("usage: dtwGenerator [options]\nGenerates a given number of sequences composed of runs of a single value, where the length of each run varies according to the given distribution, and outputs them in the format specified below. Available options");
	po::positional_options_description p;
    
    desc.add_options()
		("help,h", "output help message")
        ("no-signals,n", po::value<int>(&n)->default_value(2), "the number of sequences generated")
        ("sequence-length,l", po::value<int>(&len)->default_value(50), "the length of each sequence (i.e. the amount of runs, not the length of the \"un-RLE'd\" sequence)")
        ("dist-poisson,p", po::value<std::vector<int>>(&distributionParameters)->multitoken(), "use poisson distribution for determining block lengths (with its expected value as the input parameter)")
        ("dist-uniform,u", po::value<std::vector<int>>(&distributionParameters)->multitoken(), "use discrete uniform distribution for determining block lengths (with lower and upper bound (both inclusive) as the input parameters)")
        ("val-range,v", po::value<std::vector<int>>(&valRange)->multitoken(), "takes two integer arguments, a and b, and then draws (integer) values of the elements of the sequence from the range [a, b]");
    
    po::variables_map vm;
	po::store(po::command_line_parser(argc, argv).options(desc).positional(p).run(), vm);
	po::notify(vm);
    
    if (vm.count("help") || 
        ((!vm.count("dist-poisson")) && (!vm.count("dist-uniform")))) {
        std::cout << desc << '\n';
        std::cout << "Output format (roughly):\n(sequence-length (value, boundary)+ [newline])+\n";
        std::cout << "  e.g. 3 (1, 3) (0, 4) (-1, 8), which corresponds to the sequence (1, 1, 1, 1, 0, -1, -1, -1, -1) \n";
        std::cout << "Note: exactly one of dist-* parameters must be provided\n";
        return 0;
    }
    
    std::vector<std::vector<akt::RunI>> seqs;
    auto valDist = std::uniform_int_distribution<>(-11, 23);
    if (vm.count("val-range")) {
        if (valRange.size() < 2) {
            std::cout << "Not enough arguments for -v\n";
            return 0;
        }
        if (valRange[1] <= valRange[0]) {
            std::cout << "Invalid range for -v. The range should be [a, b], with a < b\n";
            return 0;
        }
        valDist = std::uniform_int_distribution<>(valRange[0], valRange[1]);
    }
    if (vm.count("dist-poisson"))
        seqs = akt::generateSequences(valDist, std::poisson_distribution<>(distributionParameters.front()), len, n);
    else if (vm.count("dist-uniform"))
        seqs = akt::generateSequences(valDist, std::uniform_int_distribution<>(std::max(1, distributionParameters.front()), std::max(1, distributionParameters.back())), len, n);
	else {
		std::cout << "Please provide exactly one block length distribution parameter\n";
		return 0;
	}

    for (const auto& seq : seqs) {
        std::cout << seq.size() << ' ';
        for (auto i : seq)
            std::cout << i << ' ';
        std::cout << '\n';
    }
}
