#include "intersectionPoints.h"
#include <iostream>

namespace {
    using akt::Intersections;
    using akt::IntersectionPoint;
    using akt::Position;
    using akt::Run;
    using akt::RunI;
    
    // Adds a (partially initialized) IntersectionPoint to the proper intesection
    // Note: ip has to have its loc, curBoundaries and prevPosInBoundary fields initialized
    template <typename T>
    inline void lowerLeftAddPoint(const std::vector<Run<T>>& x, const std::vector<Run<T>>& y, int xIdx, int yIdx, IntersectionPoint<T> toAdd, Intersections<T>& out)
    {
        if (x[xIdx].boundary == toAdd.loc.x) {
            toAdd.posInBoundary.x = static_cast<int>(out.xIntersections[xIdx].size());
            if (y[yIdx].boundary == toAdd.loc.y)
                toAdd.prevBelowIdx = Position{ (xIdx <= 0) ? -1 : static_cast<int>(out.xIntersections[xIdx - 1].size()), static_cast<int>(out.yIntersections[yIdx].size()) };
            out.xIntersections[xIdx].push_back(std::move(toAdd));
        }
        else {
            toAdd.posInBoundary.y = static_cast<int>(out.yIntersections[yIdx].size());
            toAdd.prevBelowIdx = Position{ (xIdx <= 0) ? -1 : static_cast<int>(out.xIntersections[xIdx - 1].size()), static_cast<int>(out.yIntersections[yIdx].size()) + 1};
            out.yIntersections[yIdx].push_back(std::move(toAdd));
        }
    }
    
    // Computes the points to the lower left of the (i, j)-th block in the run (the starting point inclusive) and adds them to out
    // Note: points which are on both x _and_ y boundary get added only to the x-intersections vector
    template <typename T>
    void getLowerLeft(const std::vector<Run<T>>& x, const std::vector<Run<T>>& y, int i, int j, Intersections<T>& out)
    {
        int xIdx = i, yIdx = j, xPos = x[i].boundary, yPos = y[j].boundary;
        while ((xIdx > 0) && (yIdx > 0)) {
            IntersectionPoint<T> toAdd;
            toAdd.loc = Position{ xPos, yPos };
            toAdd.curBoundaries = Position{ xIdx, yIdx };

            auto deltaX = xPos - x[xIdx - 1].boundary;
            auto deltaY = yPos - y[yIdx - 1].boundary;
            auto delta = std::min(deltaX, deltaY);
            if (deltaX <= deltaY)
                toAdd.prevPosInBoundary.x = static_cast<int>(out.xIntersections[xIdx - 1].size());
            else
                toAdd.prevPosInBoundary.y = static_cast<int>(out.yIntersections[yIdx - 1].size());

            lowerLeftAddPoint(x, y, xIdx, yIdx, std::move(toAdd), out);
            
            xPos -= delta;
            yPos -= delta;
            if (xPos <= x[xIdx - 1].boundary)
                --xIdx;
            if (yPos <= y[yIdx - 1].boundary)
                --yIdx;
        }

        while ((xIdx > 0) && (yPos >= 0)) {
            IntersectionPoint<T> toAdd;
            toAdd.loc = Position{ xPos, yPos };
            toAdd.curBoundaries = Position{ xIdx, yIdx };
            
            auto delta = xPos - x[xIdx - 1].boundary;
            if ((yPos - delta) >= 0) {
                toAdd.prevPosInBoundary.x = static_cast<int>(out.xIntersections[xIdx - 1].size());
            }

            lowerLeftAddPoint(x, y, xIdx, yIdx, std::move(toAdd), out);
            
            xPos -= delta;
            yPos -= delta;
            if (xPos <= x[xIdx - 1].boundary)
                --xIdx;
        }
        
        while ((yIdx > 0) && (xPos >= 0)) {
            IntersectionPoint<T> toAdd;
            toAdd.loc = Position{ xPos, yPos };
            toAdd.curBoundaries = Position{ xIdx, yIdx };
            auto delta = yPos - y[yIdx - 1].boundary;
            if ((xPos - delta) >= 0)
                toAdd.prevPosInBoundary.y = static_cast<int>(out.yIntersections[yIdx - 1].size());

            lowerLeftAddPoint(x, y, xIdx, yIdx, std::move(toAdd), out);
            
            xPos -= delta;
            yPos -= delta;
            if (yPos <= y[yIdx - 1].boundary)
                --yIdx;
        }
        
        // Add the remaining point in the (0, 0) block (if needed)
        if ((xIdx == 0) && (yIdx == 0) && (xPos >= 0) && (yPos >= 0)) {
            IntersectionPoint<T> toAdd;
            toAdd.loc = Position{ xPos, yPos };
            toAdd.curBoundaries = Position{ xIdx, yIdx };

            lowerLeftAddPoint(x, y, xIdx, yIdx, std::move(toAdd), out);
        }
    }

    // Computes the points to the lower left of the (i, j)-th block in the run (ignoring the starting point) and adds them to out
    // Note: points which are on both x _and_ y boundary get added only to the x-intersections vector
    template <typename T>
    void getUpperRight(const std::vector<Run<T>>& x, const std::vector<Run<T>>& y, int i, int j, Intersections<T>& out)
    {
        if ((i >= (x.size() - 1)) && (j >= (y.size() - 1)))
            return;
        
        int xIdx = i, yIdx = j, xPos = x[i].boundary, yPos = y[j].boundary;
        
        auto delta = (xIdx >= (x.size() - 1)) ? (y[yIdx + 1].boundary - yPos)
            : ((yIdx >= (y.size() - 1)) ? (x[xIdx + 1].boundary - xPos) : std::min(x[xIdx + 1].boundary - xPos, y[yIdx + 1].boundary - yPos));
        xPos += delta;
        yPos += delta;
        if (x[xIdx + 1].boundary <= xPos)
            ++xIdx;
        if (y[yIdx + 1].boundary <= yPos)
            ++yIdx;
        Position prevPosInBoundary = { static_cast<int>(out.xIntersections[i].size()) - 1, -1 };
        
        if ((xPos > x.back().boundary) || (yPos > y.back().boundary))
            return;
        
        while ((xIdx < (x.size() - 1)) && (yIdx < (y.size() - 1))) {
            IntersectionPoint<T> toAdd;
            toAdd.loc = Position{ xPos, yPos };
            toAdd.curBoundaries = Position{ xIdx + 1, yIdx + 1};
            toAdd.prevPosInBoundary = prevPosInBoundary;
            
            if (x[xIdx].boundary == xPos) {
                --toAdd.curBoundaries.x;
                toAdd.posInBoundary.x = static_cast<int>(out.xIntersections[xIdx].size());
                prevPosInBoundary = toAdd.posInBoundary;
                if (y[yIdx].boundary == yPos) {
                    --toAdd.curBoundaries.y;
                    toAdd.prevBelowIdx = Position{ static_cast<int>(out.xIntersections[xIdx - 1].size()), static_cast<int>(out.yIntersections[yIdx].size()) };
                }
                out.xIntersections[xIdx].push_back(std::move(toAdd));
            }
            else {
                --toAdd.curBoundaries.y;
                toAdd.posInBoundary.y = static_cast<int>(out.yIntersections[yIdx].size());
                prevPosInBoundary = toAdd.posInBoundary;
                toAdd.prevBelowIdx = Position{ static_cast<int>(out.xIntersections[xIdx].size()), static_cast<int>(out.yIntersections[yIdx].size()) + 1};
                out.yIntersections[yIdx].push_back(std::move(toAdd));
            }
            
            auto deltaX = x[xIdx + 1].boundary - xPos;
            auto deltaY = y[yIdx + 1].boundary - yPos;
            auto delta = std::min(deltaX, deltaY);
            
            xPos += delta;
            yPos += delta;
            if (x[xIdx + 1].boundary <= xPos)
                ++xIdx;
            if (y[yIdx + 1].boundary <= yPos)
                ++yIdx;
        }
        
        // Add the remaining point on the edge 
        xIdx = std::max(xIdx, 0);
        yIdx = std::max(yIdx, 0);
        IntersectionPoint<T> toAdd;
        toAdd.loc = Position{ xPos, yPos };
        toAdd.curBoundaries = Position{ xIdx, yIdx };
        toAdd.prevPosInBoundary = prevPosInBoundary;
        if (xPos > x[xIdx].boundary)
            ++toAdd.curBoundaries.x;
        if (yPos > y[yIdx].boundary)
            ++toAdd.curBoundaries.y;

        if (xPos == x[xIdx].boundary) {
            toAdd.posInBoundary = { static_cast<int>(out.xIntersections[xIdx].size()), -1 };
            if (y[yIdx].boundary == yPos) // Corner point
                toAdd.prevBelowIdx = Position{ static_cast<int>(out.xIntersections[xIdx - 1].size()), static_cast<int>(out.yIntersections[yIdx].size()) };
            out.xIntersections[xIdx].push_back(std::move(toAdd));
        }
        else {
            toAdd.posInBoundary.y = static_cast<int>(out.yIntersections[yIdx].size());
            toAdd.prevBelowIdx = Position{ static_cast<int>(out.xIntersections[xIdx].size()), static_cast<int>(out.yIntersections[yIdx].size()) + 1 };
            out.yIntersections[yIdx].push_back(std::move(toAdd));
        }
    }

    // Similar to getUpperRight, but initializes the cost at first point
    template <typename T>
    void startingDiagonal(const std::vector<Run<T>>& x, const std::vector<Run<T>>& y, Intersections<T>& out)
    {
        int xIdx = -1, yIdx = -1, xPos = 0, yPos = 0;
        
        auto delta = std::min(x.front().boundary, y.front().boundary);
        xPos += delta;
        yPos += delta;
        if (x[xIdx + 1].boundary <= xPos)
            ++xIdx;
        if (y[yIdx + 1].boundary <= yPos)
            ++yIdx;
        Position prevPosInBoundary = { -1, -1 };
        
        while ((xIdx < (static_cast<int>(x.size()) - 1)) && (yIdx < (static_cast<int>(y.size()) - 1))) {
            IntersectionPoint<T> toAdd;
            toAdd.loc = Position{ xPos, yPos };
            toAdd.curBoundaries = Position{ xIdx + 1, yIdx + 1};
            toAdd.prevPosInBoundary = prevPosInBoundary;
            // Initialize cost for the starting point
            // Only the starting point has no predecessor
            if ((prevPosInBoundary.x == -1) && (prevPosInBoundary.y == -1))
                toAdd.cost = (delta + 1) * (x.front().val - y.front().val) * (x.front().val - y.front().val);
            
            if ((xIdx >= 0) && (x[xIdx].boundary == xPos)) {
                --toAdd.curBoundaries.x;
                toAdd.posInBoundary = { static_cast<int>(out.xIntersections[xIdx].size()), -1 };
                prevPosInBoundary = toAdd.posInBoundary;
                if (y[yIdx].boundary == yPos) {
                    --toAdd.curBoundaries.y;
                    toAdd.prevBelowIdx = Position{(xIdx <= 0) ? -1 : static_cast<int>(out.xIntersections[xIdx - 1].size()), static_cast<int>(out.yIntersections[yIdx].size()) };
                }
                out.xIntersections[xIdx].push_back(std::move(toAdd));
            }
            else {
                --toAdd.curBoundaries.y;
                toAdd.posInBoundary.y = static_cast<int>(out.yIntersections[yIdx].size());
                prevPosInBoundary = toAdd.posInBoundary;
                toAdd.prevBelowIdx = Position{ (xIdx < 0) ? -1 : static_cast<int>(out.xIntersections[xIdx].size()), static_cast<int>(out.yIntersections[yIdx].size()) + 1};
                out.yIntersections[yIdx].push_back(std::move(toAdd));
            }
            
            auto deltaX = x[xIdx + 1].boundary - xPos;
            auto deltaY = y[yIdx + 1].boundary - yPos;
            auto delta = std::min(deltaX, deltaY);
            xPos += delta;
            yPos += delta;
            if (x[xIdx + 1].boundary <= xPos)
                ++xIdx;
            if (y[yIdx + 1].boundary <= yPos)
                ++yIdx;
        }
        
        // Add the remaining point on the edge 
        xIdx = std::max(xIdx, 0);
        yIdx = std::max(yIdx, 0);
        IntersectionPoint<T> toAdd;
        toAdd.loc = Position{ xPos, yPos };
        toAdd.curBoundaries = Position{ xIdx, yIdx };
        toAdd.prevPosInBoundary = prevPosInBoundary;
        if (xPos > x[xIdx].boundary)
            ++toAdd.curBoundaries.x;
        if (yPos > y[yIdx].boundary)
            ++toAdd.curBoundaries.y;
        // Initialize cost for the starting point
        // Only the starting point has no predecessor
        if ((prevPosInBoundary.x == -1) && (prevPosInBoundary.y == -1))
            toAdd.cost = (delta + 1) * (x.front().val - y.front().val) * (x.front().val - y.front().val);

        if (xPos == x[xIdx].boundary) {
            toAdd.posInBoundary.x = static_cast<int>(out.xIntersections[xIdx].size());
            if (y[yIdx].boundary == yPos) // Corner point
                toAdd.prevBelowIdx = Position{ (xIdx <= 0) ? -1 : static_cast<int>(out.xIntersections[xIdx - 1].size()), static_cast<int>(out.yIntersections[yIdx].size()) };
            out.xIntersections[xIdx].push_back(std::move(toAdd));
        }
        else {
            toAdd.posInBoundary.y = static_cast<int>(out.yIntersections[yIdx].size());
            toAdd.prevBelowIdx = Position{ static_cast<int>(out.xIntersections[xIdx].size()), static_cast<int>(out.yIntersections[yIdx].size()) + 1 };
            out.yIntersections[yIdx].push_back(std::move(toAdd));
        }
    }

    // Computes all the intersection points lying on the diagonal corresponding to the (i, j)-th block in the run and add them to out
    template <typename T>
    void getIntersectionPointsOnDiag(const std::vector<Run<T>>& x, const std::vector<Run<T>>& y, int i, int j, Intersections<T>& out)
    {
        getLowerLeft(x, y, i, j, out);
        getUpperRight(x, y, i, j, out);
    }

    // Returns a list of diagonals including the start diagonal (may contain duplicates)
    template <typename T>
    inline std::vector<Position> getDiagonals(const std::vector<Run<T>>& x, const std::vector<Run<T>>& y)
    {
        std::vector<Position> res;
        res.reserve(x.size()*y.size());
        for (int i = 0; i < x.size(); ++i) {
            for (int j = 0; j < y.size(); ++j)
                res.push_back(Position{i, j});
        }
        return res;
    }
    
} // end anonymous namespace 

namespace akt {

	// Returns the total number of intersection points contained in intersections
	template <typename T>
	int64_t kappa(const Intersections<T>& intersections)
	{
		int64_t res = 0;

		res += std::accumulate(intersections.xIntersections.cbegin(), intersections.xIntersections.cend(), static_cast<size_t>(0), [](size_t acc, const auto& vec) {return acc + vec.size(); });
		res += std::accumulate(intersections.yIntersections.cbegin(), intersections.yIntersections.cend(), static_cast<size_t>(0), [](size_t acc, const auto& vec) {return acc + vec.size(); });

		return res;
	}

	// Returns the diagonals (in the form of indices into x and y vectors) listed in the upper left-lower right order (may contain duplicates)
	// Note: may not include the diagonal through the point (0, 0) (if it's not a block corner)
	template <typename T>
	std::vector<Position> getOrderedDiagonals(const std::vector<Run<T>>& x, const std::vector<Run<T>>& y)
	{
		auto diags = getDiagonals(x, y);
		std::sort(diags.begin(), diags.end(),
			[&](const auto& p1, const auto& p2) {return (x[p1.x].boundary - y[p1.y].boundary) > (x[p2.x].boundary - y[p2.y].boundary); });

		return diags;
	}

	// Returns all intersection points relevant to DTW computation
	template <typename T>
	Intersections<T> getIntersectionPoints(const std::vector<Run<T>>& x, const std::vector<Run<T>>& y)
	{
		auto diags = getOrderedDiagonals(x, y);
		// Remove duplicates from diags
		diags.erase(std::unique(diags.begin(), diags.end(),
			[&](const auto& p1, const auto& p2) {return (x[p1.x].boundary - y[p1.y].boundary) == (x[p2.x].boundary - y[p2.y].boundary); }),
			diags.end());

		auto res = Intersections<T>(x.size(), y.size());
		// Handle diagonals preceeding the one going through (0, 0)
		int i = 0;
		for (; i < diags.size(); ++i) {
			if (x[diags[i].x].boundary - y[diags[i].y].boundary == 0) {
				// Ignore the diagonal through (0, 0) as this one needs special handling
				++i;
				break;
			}
			else if (x[diags[i].x].boundary - y[diags[i].y].boundary < 0) {
				break;
			}
			getIntersectionPointsOnDiag(x, y, diags[i].x, diags[i].y, res);
		}

		// Handle the start diagonal separately
		startingDiagonal(x, y, res);

		// Handle the rest of diagonals
		for (; i < diags.size(); ++i)
			getIntersectionPointsOnDiag(x, y, diags[i].x, diags[i].y, res);

		return res;
	}
	
	// Needed for the code to link (a bit of a hack but we only use ints, int64_t's and doubles, anyway
	template int64_t kappa<int>(const Intersections<int>&);
    template int64_t kappa<int64_t>(const Intersections<int64_t>&);
    template int64_t kappa<double>(const Intersections<double>&);
	
	template std::vector<Position> getOrderedDiagonals<int>(const std::vector<Run<int>>&, const std::vector<Run<int>>&);
    template std::vector<Position> getOrderedDiagonals<int64_t>(const std::vector<Run<int64_t>>&, const std::vector<Run<int64_t>>&);
    template std::vector<Position> getOrderedDiagonals<double>(const std::vector<Run<double>>&, const std::vector<Run<double>>&);
    
    template Intersections<int> getIntersectionPoints<int>(const std::vector<Run<int>>&, const std::vector<Run<int>>&);
    template Intersections<int64_t> getIntersectionPoints<int64_t>(const std::vector<Run<int64_t>>&, const std::vector<Run<int64_t>>&);
    template Intersections<double> getIntersectionPoints<double>(const std::vector<Run<double>>&, const std::vector<Run<double>>&);

} // end namespace akt
