#pragma once

#include <algorithm>
#include <cstdint>

#include "utility.h"

namespace akt {
    template <typename T>
	struct IntersectionPoint {
		// Absolute location on the grid (as in, in the non-RLE'd signal)
		Position loc = Position{ -1, -1 };
		// Indices of the first boundary at a greater-than-or-equal-to position
		Position curBoundaries = Position{ -1, -1 };
		// Indices into the vector of points at each of the boundaries
		// Note: at least one of those will be equal to -1 (depending on whether the point lies on an x- or y-boundary)
		Position posInBoundary = Position{ -1, -1 };
		// Indices into the vector of points at each of the boundaries for the diagonally preceeding point
		// Note: at least one of those will always be equal to -1 (depending on whether the next closest point lies on x- or y-boundary)
		Position prevPosInBoundary = Position{ -1, -1 };
		// Index into the proper vector containing the point directly below the current point
		// At most one of those entries is represents the actual; the other will may equal -1 or be equal to the size of the given vector or be below the correct point; first entry may be one too small
		Position prevBelowIdx = Position{ -1, -1 };
		// Cost at the current point
		T cost = -1;
	};

	// Stores the intersections between diagonal and horizontal/vertical lines
	// xIntersections stores the intersections between the diagonals and the x-lines, in sorted order for each of the lines
	// For now the containers are ordered maps/sets, perhaps can change (some of) them to normal vectors for speed
    template <typename T>
	struct Intersections
	{
		Intersections(int xSz, int ySz)
			: xIntersections(xSz), yIntersections(ySz) { }

		std::vector<std::vector<IntersectionPoint<T>>> xIntersections;
		std::vector<std::vector<IntersectionPoint<T>>> yIntersections;
	};

	// Returns the total number of intersection points contained in intersections
    template <typename T>
	int64_t kappa(const Intersections<T>& intersections);

	// Returns the diagonals (in the form of indices into x and y vectors to a point on that diagonal) listed in the upper left-lower right order (may contain duplicates)
	// Note: may not include the diagonal through the point (0, 0) (if it's not a block corner)
    template <typename T>
	std::vector<Position> getOrderedDiagonals(const std::vector<Run<T>>& x, const std::vector<Run<T>>& y);

	// Returns all intersection points relevant to DTW computation
    template <typename T>
	Intersections<T> getIntersectionPoints(const std::vector<Run<T>>& x, const std::vector<Run<T>>& y);

} // end namespace akt
