#include "sparsedtw.h"
#include <iostream>

// Note: here the axes are in the usual (in maths) order (unlike in implementations of some other algorithms in this project):
//  the (0,0) point is in the bottom left-hand corner and xs increase to the right

using akt::SparseObservation;

namespace {
    // Stores the DP vectores as one entity
    // Note: corner points are stored in both x vectors and y vectors!
    template <typename T>
    struct DpPoints {
        DpPoints(size_t noXs, size_t lastXpos, size_t noYs, size_t lastYpos)
            : x(noXs, std::vector<T>(lastYpos + 1)), y(noYs, std::vector<T>(lastXpos + 1)) { }
        // Lines of constant x
        std::vector<std::vector<T>> x;
        // Lines of constant y
        std::vector<std::vector<T>> y;
    };
    
    // Initializes the first (i.e. bottom right, or (0, 0)) rectangle for the SparseDTW DP table
    template <typename T>
    void sparseInitializeBottomLeft(const std::vector<SparseObservation<T>>& x, const std::vector<SparseObservation<T>>& y, DpPoints<T>& out)
    {
        if (x[0].pos == 0) {
            if (y[0].pos == 0) {
                // Neight right nor top edge exists, so the only cell to initialize is the top-right corner
                out.x[0][y[0].pos] = out.y[0][x[0].pos] = (x[0].val - y[0].val) * (x[0].val - y[0].val);
            } else {
                // Only the right edge exists; initlize it properly (note: this is a different initialization than in the case where both edges exist!)
                auto rightCost = x[0].val * x[0].val;
                out.x[0][0] = rightCost;
                for (int j = 1; j < y[0].pos; ++j)
                    out.x[0][j] = out.x[0][j - 1] + rightCost;
                out.x[0][y[0].pos] = out.y[0][x[0].pos] = out.x[0][y[0].pos - 1] + (x[0].val - y[0].val) * (x[0].val - y[0].val);
            }   
        } else if (y[0].pos == 0) {
            // Only the top edge exists; initlize it properly (note: this is a different initialization than in the case where both edges exist!)
            auto topCost = y[0].val * y[0].val;
            out.y[0][0] = topCost;
            for (int i = 1; i < x[0].pos; ++i)
                out.y[0][i] = out.y[0][i - 1] + topCost;
            out.x[0][y[0].pos] = out.y[0][x[0].pos] = out.y[0][x[0].pos - 1] + (x[0].val - y[0].val) * (x[0].val - y[0].val);
        } else {
            // Both edges exist; initialize them
            // Right edge:
            auto rightCost = x[0].val * x[0].val;
            for (int j = 0; j < y[0].pos; ++j)
                out.x[0][j] = rightCost;
            // Top edge:
            auto topCost = y[0].val * y[0].val;
            for (int i = 0; i < x[0].pos; ++i)
                out.y[0][i] = topCost;
            // Top right corner (V_r,t using the notation from the paper)
            out.x[0][y[0].pos] = out.y[0][x[0].pos] = (x[0].val - y[0].val) * (x[0].val - y[0].val);
        }
    }
    
    // Initializes all rectangles directly above the first one (i.e. those with indices (0, k) for k > 0)
    template <typename T>
    void sparseInitializeTop(const std::vector<SparseObservation<T>>& x, const std::vector<SparseObservation<T>>& y, DpPoints<T>& out)
    {
        auto rightCost = x[0].val * x[0].val;
        if (x[0].pos == 0) {
            // The trivial case: we only have one line going to the top
            for (int i = 1; i < y.size(); ++i) {
                // Initialize the edge
                for (int j = y[i - 1].pos + 1; j < y[i].pos; ++j)
                    out.x[0][j] = out.x[0][j - 1] + rightCost;
                // Initialize the corner
                out.x[0][y[i].pos] = out.y[i][x[0].pos] = 
                    (x[0].val - y[i].val) * (x[0].val - y[i].val) +
                    out.x[0][y[i].pos - 1];
            }
            return;
        }
        
        // The non-trivial case: all rectangles have a top edge
        for (int i = 1; i < y.size(); ++i) {
            auto blockCost = out.y[i - 1][0]; // The cost on the inside of the block
            // Initialize the top edge
            auto topCost = y[i].val * y[i].val;
            out.y[i][0] = blockCost + topCost;
            if (y[i].pos > (y[i - 1].pos + 1)) {
                for (int j = 1; j < x[0].pos; ++j)
                    out.y[i][j] = out.y[i][0];
            } else {
                for (int j = 1; j < x[0].pos; ++j)
                    out.y[i][j] = rightCost + out.y[i - 1][j - 1];
            }
            // Initialize the right edge (if it exists)
            for (int j = y[i - 1].pos + 1; j < y[i].pos; ++j)
                out.x[0][j] = rightCost + std::min(blockCost, out.x[0][j - 1]);
            // Initialize the top-right corner
            out.x[0][y[i].pos] = out.y[i][x[0].pos] = 
                (x[0].val - y[i].val) * (x[0].val - y[i].val) + 
                std::min(blockCost, out.x[0][y[i].pos - 1]);
                
            out.x[0][y[i].pos] = out.y[i][x[0].pos] = 
                (x[0].val - y[i].val) * (x[0].val - y[i].val);
            auto diff = std::min(blockCost, out.x[0][y[i].pos - 1]);
            if (y[i].pos == (y[i - 1].pos + 1))
                diff = std::min(out.y[i - 1][x[0].pos], out.y[i - 1][x[0].pos - 1]);
            out.x[0][y[i].pos] += diff;
            out.y[i][x[0].pos] += diff;
        }
    }
    
    // Initializes all rectangles directly to the right of the first one (i.e. those with indices (k, 0) for k > 0)
    template <typename T>
    void sparseInitializeRight(const std::vector<SparseObservation<T>>& x, const std::vector<SparseObservation<T>>& y, DpPoints<T>& out)
    {
        auto topCost = y[0].val * y[0].val;
        if (y[0].pos == 0) {
            // The trivial case: we only have one line going to the right
            for (int i = 1; i < x.size(); ++i) {
                // Initialize the edge
                for (int j = x[i - 1].pos + 1; j < x[i].pos; ++j)
                    out.y[0][j] = out.y[0][j - 1] + topCost;
                // Initialize the corner
                out.y[0][x[i].pos] = out.x[i][y[0].pos] = 
                    (x[i].val - y[0].val) * (x[i].val - y[0].val) +
                    out.y[0][x[i].pos - 1];
            }
            return;
        }
        
        // The non-trivial case: all rectangles have a right edge
        for (int i = 1; i < x.size(); ++i) {
            auto blockCost = out.x[i - 1][0]; // The cost on the inside of the block
            // Initialize the right edge
            auto rightCost = x[i].val * x[i].val;
            out.x[i][0] = blockCost + rightCost;
            if (x[i].pos > (x[i - 1].pos + 1)) {
                for (int j = 1; j < y[0].pos; ++j)
                    out.x[i][j] = out.x[i][0];
            } else {
                for (int j = 1; j < y[0].pos; ++j)
                    out.x[i][j] = rightCost + out.x[i - 1][j - 1];
            }
            
            // Initialize the top edge (if it exists)
            for (int j = x[i - 1].pos + 1; j < x[i].pos; ++j)
                out.y[0][j] = topCost + std::min(blockCost, out.y[0][j - 1]);
            // WRONG (CHANGE): Initialize the top-right corner; notice that since the right edge is always greater than blockCost, we can only come in diagonally or from the left
            out.y[0][x[i].pos] = out.x[i][y[0].pos] = 
                (x[i].val - y[0].val) * (x[i].val - y[0].val);
            auto diff = std::min(blockCost, out.y[0][x[i].pos - 1]);
            if (x[i].pos == (x[i - 1].pos + 1))
                diff = std::min(out.x[i - 1][y[0].pos], out.x[i - 1][y[0].pos - 1]);
            out.y[0][x[i].pos] += diff;
            out.x[i][y[0].pos] += diff;    
        }
    }
    
    // Returns an initialized DP table for all entries of type (0,k) and (k,0)
    // First element of the pair contains the lines of constant x (and so it has dimensions x.size() by y.back().pos),
    //  and the second has lines of constant y (and so it has dimensions y.size() by x.back().pos)
    template <typename T>
    DpPoints<T> sparseInitializeDp(const std::vector<SparseObservation<T>>& x, const std::vector<SparseObservation<T>>& y)
    {
        auto res = DpPoints<T>(x.size(), x.back().pos, y.size(), y.back().pos);
        
        // Initialize the DP array in three pieces
        sparseInitializeBottomLeft(x, y, res);
        sparseInitializeTop(x, y, res);
        sparseInitializeRight(x, y, res);
        
        return res;
    }
    
    // Handles the trivial case (the other signal consists of only zeros)
    template <typename T>
    T handleTrivial(const std::vector<SparseObservation<T>>& x)
    {
        return std::accumulate(x.cbegin(), x.cend(), static_cast<T>(0),
                               [](auto acc, auto so) {return acc + so.val * so.val;});
    }
    
} // end anonymous namespace

namespace akt {
    
    // Converts a "flat" signal to the SparseObservation format
    template <typename T>
    std::vector<SparseObservation<T>> toSparseObservation(const std::vector<T>& x)
    {
        // Almost a copy-paste of the toRLE function (except for checking whether a potential run is of zeros only)
        // Note: this "compression" is slightly lossy (if x ends with a run of zeros, the run will be truncated)
		std::vector<SparseObservation<T>> res;
		for (int i = 0; i < x.size(); ++i) {
			if (x[i] != static_cast<T>(0))
                res.emplace_back(x[i], i);
		}

		return res;
	}
    
    // Runs the SparseDTW (SDTW) algorithm on the given sequence and returns the computed DTW distance
    template <typename T>
    T sparseDtw(const std::vector<SparseObservation<T>>& x, const std::vector<SparseObservation<T>>& y)
    {
        if (x.size() == 0)
            return handleTrivial(y);
        if (y.size() == 0)
            return handleTrivial(x);
        
        auto dp = sparseInitializeDp(x, y);
        
        for (int i = 1; i < x.size(); ++i) {
            for (int j = 1; j < y.size(); ++j) {
                auto blockCost = std::min(dp.x[i - 1][y[j - 1].pos],
                    std::min(dp.x[i - 1][y[j - 1].pos + 1], dp.y[j - 1][x[i - 1].pos + 1]));
                auto rightCost = x[i].val * x[i].val;
                auto topCost = y[j].val * y[j].val;
                auto cornerCost = (x[i].val - y[j].val) * (x[i].val - y[j].val);
                // Handle the 4 different cases (the normal case and 3 degenerate ones)
                if (x[i].pos == (x[i - 1].pos + 1)) {
                    if (y[j].pos == (y[j - 1].pos + 1)) {
                        // Most degenerate case; basically standard DTW
                        dp.x[i][y[j].pos] = dp.y[j][x[i].pos] =
                            cornerCost + blockCost;
                    } else {
                        // No top edge, but right edge exists
                        dp.x[i][y[j - 1].pos + 1] = rightCost + blockCost;
                        for (int k = y[j - 1].pos + 2; k < y[j].pos; ++k)
                            dp.x[i][k] = rightCost + std::min(dp.x[i - 1][k - 1], dp.x[i][k - 1]); // Include left?
                        dp.x[i][y[j].pos] = dp.y[j][x[i].pos] =
                            cornerCost + std::min(dp.x[i - 1][y[j].pos - 1],
                            std::min(dp.x[i - 1][y[j].pos], dp.x[i][y[j].pos - 1]));
                    }
                } else if (y[j].pos == (y[j - 1].pos + 1)) {
                    // No right edge, but top edge exists
                    dp.y[j][x[i - 1].pos + 1] = topCost + blockCost;
                    for (int k = x[i - 1].pos + 2; k < x[i].pos; ++k)
                        dp.y[j][k] = topCost + std::min(dp.y[j - 1][k - 1], dp.y[j][k - 1]); // Include below?
                    dp.x[i][y[j].pos] = dp.y[j][x[i].pos] =
                        cornerCost + std::min(dp.y[j - 1][x[i].pos - 1],
                        std::min(dp.y[j - 1][x[i].pos], dp.y[j][x[i].pos - 1]));
                } else {
                    // Normal case: all edges exist and there exists an interior of the block
                    // Right edge:
                    dp.x[i][y[j - 1].pos + 1] = rightCost + std::min(blockCost, dp.x[i][y[j - 1].pos]); // Include diagonal? Join it with the loop below?
                    for (int k = y[j - 1].pos + 2; k < y[j].pos; ++k)
                        dp.x[i][k] = rightCost + std::min(blockCost, dp.x[i][k - 1]);
                    // Top edge:
                    dp.y[j][x[i - 1].pos + 1] = topCost + std::min(blockCost, dp.y[j][x[i - 1].pos]); // Include diagonal? Join it with the loop below?
                    for (int k = x[i - 1].pos + 2; k < x[i].pos; ++k)
                        dp.y[j][k] = topCost + std::min(blockCost, dp.y[j][k - 1]);
                    // Top-right corner
                    dp.x[i][y[j].pos] = dp.y[j][x[i].pos] =
                        cornerCost + std::min(blockCost,
                        std::min(dp.x[i][y[j].pos - 1], dp.y[j][x[i].pos - 1]));
                }
            }
        }
        
        return dp.x.back().back();
    }
    
    // The below is so that the code links (a bit of a hack but we only use ints, int64_t's and doubles anyway
	template std::vector<SparseObservation<int>> toSparseObservation<int>(const std::vector<int>&);
    template std::vector<SparseObservation<int64_t>> toSparseObservation<int64_t>(const std::vector<int64_t>&);
    template std::vector<SparseObservation<double>> toSparseObservation<double>(const std::vector<double>&);
    
    template int sparseDtw<int>(const std::vector<SparseObservation<int>>&, const std::vector<SparseObservation<int>>&);
    template int64_t sparseDtw<int64_t>(const std::vector<SparseObservation<int64_t>>&, const std::vector<SparseObservation<int64_t>>&);
    template double sparseDtw<double>(const std::vector<SparseObservation<double>>&, const std::vector<SparseObservation<double>>&);
    
} // end namespace akt
