#include "utility.h"

namespace akt {
	std::random_device GLOBAL_rd;
	std::mt19937 GLOBAL_gen(GLOBAL_rd());
}

namespace akt {
	// Converts a sequence to our Run Length Encoding
	template <typename T>
	std::vector<Run<T>> toRLE(const std::vector<T>& seq)
    {
		std::vector<Run<T>> res;
		for (int i = 0; i < seq.size();) {
			// Compute the length of the run
			int j = i + 1;
			while ((j < seq.size()) && (seq[j] == seq[i]))
				++j;
			// Add the run to the result and start the next run
			res.emplace_back(seq[i], j - 1);
			i = j;
		}

		return res;
	}

	// Converts a sequence from our RLE to a serial form
    template <typename T>
	std::vector<T> fromRLE(const std::vector<Run<T>>& rle)
    {
        if (rle.size() == 0)
            return std::vector<T>();
		std::vector<T> res;
        res.reserve(rle.back().boundary + 1);

        res.insert(res.end(), rle.front().boundary + 1, rle.front().val);
		for (int i = 1; i < rle.size(); ++i)
			res.insert(res.end(), rle[i].boundary - rle[i - 1].boundary, rle[i].val);
            
		return res;
	}

	// Outputs the run in a (val, boundary) format
    template <typename T>
	std::ostream& operator<<(std::ostream& os, const Run<T>& r)
    {
		os << '(' << r.val << ", " << r.boundary << ')';
		return os;
	}

	// Reads in a run in a (val, boundary) format
	// Note: may consume some characters from is on failure, so use only if its acceptable
	template <typename T>
    std::istream& operator>>(std::istream& is, Run<T>& r)
	{
		char openPar, comma, closingPar;
		decltype(r.val) val;
		decltype(r.boundary) boundary;
		is >> openPar >> val >> comma >> boundary >> closingPar;
		if (!is)
			return is;
		if ((openPar != '(') || (comma != ',') || (closingPar != ')')) {
			is.clear(std::ios_base::failbit);
			return is;
		}
		r = Run(val, boundary);

		return is;
	}

	// Outputs the number of Runs in the vector and then the runs
	// e.g. "3 (1, 3) (0, 4) (-1, 8)"
    template <typename T>
	void outputRuns(std::ostream& os, const std::vector<Run<T>>& x)
	{
		os << x.size();
		for (const auto& r : x)
			os << ' ' << r;
	}

	// Reads in the amount of Runs (separated by whitespace) to be read and then reads them
	// Note: Assumes the input is valid
    template <typename T>
	std::optional<std::vector<Run<T>>> readRuns(std::istream& is)
	{
		int runs;
		if (!(is >> runs))
            return {};
        
		std::vector<Run<T>> res;
		res.reserve(runs);
		for (int i = 0; i < runs; ++i) {
			Run<T> r(static_cast<T>(-1), -1);
			is >> r;
			res.push_back(std::move(r));
		}

		return res;
	}
	
	
	// The below is so that the code links (a bit of a hack but we only use Run<int> Run<int64_t> and Run<double> anyway
	template std::vector<Run<int>> toRLE<int>(const std::vector<int>&);
    template std::vector<Run<int64_t>> toRLE<int64_t>(const std::vector<int64_t>&);
    template std::vector<Run<double>> toRLE<double>(const std::vector<double>&);
	
	template std::vector<int> fromRLE<int>(const std::vector<Run<int>>&);
    template std::vector<int64_t> fromRLE<int64_t>(const std::vector<Run<int64_t>>&);
    template std::vector<double> fromRLE<double>(const std::vector<Run<double>>&);
	
	template std::ostream& operator<<<int>(std::ostream&, const Run<int>&);
    template std::ostream& operator<<<int64_t>(std::ostream&, const Run<int64_t>&);
    template std::ostream& operator<<<double>(std::ostream&, const Run<double>&);
	
	template std::istream& operator>><int>(std::istream&, Run<int>&);
    template std::istream& operator>><int64_t>(std::istream&, Run<int64_t>&);
    template std::istream& operator>><double>(std::istream&, Run<double>&);
	
	template void outputRuns<int>(std::ostream&, const std::vector<Run<int>>&);
    template void outputRuns<int64_t>(std::ostream&, const std::vector<Run<int64_t>>&);
    template void outputRuns<double>(std::ostream&, const std::vector<Run<double>>&);
	
    template std::optional<std::vector<Run<int>>> readRuns<int>(std::istream&);
	template std::optional<std::vector<Run<int64_t>>> readRuns<int64_t>(std::istream&);
    template std::optional<std::vector<Run<double>>> readRuns<double>(std::istream&);
    
} // end namespace akt
