#pragma once

#include <cstdint>
#include <iostream>
#include <limits>
#include <optional>
#include <random>
#include <vector>

namespace akt {
    template <typename T>
    constexpr T COST_INFINITE = std::numeric_limits<T>::max();

	extern std::random_device GLOBAL_rd;
	extern std::mt19937 GLOBAL_gen;
    
    template <typename T>
	struct Run
	{
        Run(T val, int boundary)
			: val(val), boundary(boundary) { }

		// The value saved
		T val;
		// The top boundary of the run in the original sequence, i.e. the run spans the range [boundary - (length of the run) + 1, boundary]
		int boundary;
	};

    using RunI = Run<int>;
    
	// Outputs the run in a (val, boundary) format
    template <typename T>
	std::ostream& operator<<(std::ostream& os, const Run<T>& r);

	// Reads in a run in a (val, boundary) format
	// Note: may consume some characters from is on failure, so use only if that behavior is deemed acceptable
	template <typename T>
    std::istream& operator>>(std::istream& is, Run<T>& r);

	// Outputs the number of Runs in the vector and then the runs
	// e.g. "3 (1, 3) (0, 4) (-1, 8)"
    template <typename T>
	void outputRuns(std::ostream& os, const std::vector<Run<T>>& x);

	// Reads in the amount of Runs (separated by whitespace) to be read and then reads them
    // If stream reached EOF, no vector is contained in the optional
	// Note: Assumes the input is valid
    template <typename T>
	std::optional<std::vector<Run<T>>> readRuns(std::istream& is);

	struct Position {
		int x = -1;
		int y = -1;
	};

	inline bool operator==(const Position& p1, const Position& p2)
	{
		return (p1.x == p2.x) && (p1.y == p2.y);
	}
	inline bool operator!=(const Position& p1, const Position& p2)
	{
		return !(p1 == p2);
	}

	// Converts a sequence to our Run Length Encoding
	template <typename T>
	std::vector<Run<T>> toRLE(const std::vector<T>& seq);

	// Converts a sequence from our RLE to a serial form
    template <typename T>
	std::vector<T> fromRLE(const std::vector<Run<T>>& rle);

	// Returns a randomly generated sequence of runs, where values and lengths of the runs are taken from the given distributions (using the ones defined in <random>)
	// Note: if a run has length <= 0 it gets ignored (it is "rerolled")
	// Note: if a value at a run is the same as the previous block, the value gets "rerolled"
	template <class ValueDist, class RunLenDist>
	std::vector<Run<decltype(ValueDist()(GLOBAL_gen))>> randomSequence(ValueDist vd, RunLenDist rd, int noRuns)
	{
		if (noRuns <= 0)
			return std::vector<Run<decltype(ValueDist()(GLOBAL_gen))>>();

		std::vector<Run<decltype(ValueDist()(GLOBAL_gen))>> res;
		res.reserve(noRuns);

		int curBoundary = -1;
		for (int i = 0; i < noRuns; ++i) {
			auto val = vd(GLOBAL_gen);
			auto len = -1;
			while ((len = rd(GLOBAL_gen)) <= 0)
				;
            while ((res.size() > 0) && (val == res.back().val))
                val = vd(GLOBAL_gen);
            
			curBoundary += len;
			res.emplace_back(val, curBoundary);
		}

		return res;
	}

} // end namespace akt
