
#pragma once

#include "utils.hpp"
#include "block.hpp"
#include<vector>
#include<list>
#include <string>

namespace kappa {
  std::string class_to_str(const int c)
  {
    std::string result;
    if(c & vertical) result += "vertical ";
    if(c & horizontal) result += "horizontal ";
    if(c & diagonal) result += "diagonal ";
    return result;
  }

  struct RawDiagonal
  {
    const int_fast32_t x_offset;

    RawDiagonal(const int_fast32_t _offset): x_offset(_offset) {}
  };


  inline bool operator<(const RawDiagonal& d, const RawBlock& b)  { return b.x_offset() > d.x_offset; }
  inline bool operator<=(const RawDiagonal& d, const RawBlock& b)  { return b.x_offset() >= d.x_offset; }
  inline bool operator==(const RawDiagonal& d, const RawBlock& b) { return b.x_offset() == d.x_offset; }
  inline bool operator!=(const RawDiagonal& d, const RawBlock& b) { return b.x_offset() != d.x_offset; }


  template<class T = double>
  struct Diagonal: public RawDiagonal
  {
    using RawDiagonal::x_offset;
    //using CellStorage = std::list<Cell>;
    using CellStorage = std::vector<Cell<T>>;
    using RevCellIter = typename std::reverse_iterator<typename CellStorage::const_iterator>;

    // where the diagonal intersects the y=0 line
    const Cell<T>* last_added_cell = nullptr;

    // keep track of number of cells added to this diagonal
    STAT(uint_fast32_t kappa = 0);

    // relevant cells on the diagonal, starting from bottom; {vert,horiz} contains all intersection with {vertical,horizontal} boundaries, indexed by {i,j}
    CellStorage relevant_cells_vert;
    CellStorage relevant_cells_hori;

    Diagonal(const int_fast32_t _offset, const uint_fast32_t runs_x, const uint_fast32_t runs_y): 
      RawDiagonal(_offset)
    {
      relevant_cells_vert.reserve(runs_x);
      relevant_cells_hori.reserve(runs_y);
    }

    template<int target>
    inline void add_cell(const int_fast32_t x, const int_fast32_t y, const T cost)
    {
      if(target & (diagonal | horizontal)) last_added_cell = &(relevant_cells_hori.emplace_back(x, y, cost));
      if(target & (diagonal | vertical)) last_added_cell = &(relevant_cells_vert.emplace_back(x, y, cost));
      DEBUG(std::cout << "\tadded "<<class_to_str(target)<<" cell "<<*last_added_cell<<" to diagonal "<<x_offset<<'\n');
      STAT(++kappa);
    }

    template<int target, int cost_from>
    inline void add_cell(const Block<T>& b, const Cell<T>* hori_from, const Cell<T>* vert_from, const int_fast32_t x, const int_fast32_t y)
    {
      DEBUG(std::cout << "adding cell ("<<x<<","<<y<<") intersecting block with x_offset "<<b.x_offset()<<" & cost "<<b.cost<<" for target "<<target<<" and cost from "<<cost_from<<" - " << class_to_str(cost_from)<<'\n');

      assert((cost_from > 0) && (cost_from <= diagonal + horizontal + vertical));
      assert((target > 0) && (target <= diagonal + horizontal + vertical));
      
      T cost = INT32_MAX;
      if((cost_from & diagonal) && (last_added_cell != nullptr)) cost = cost_to<diagonal>(x, y, b.cost);
      if(cost_from & horizontal) cost = std::min(cost, hori_from->horizontal_cost(x, b.cost));
      if(cost_from & vertical) cost = std::min(cost, vert_from->vertical_cost(y, b.cost));
      add_cell<target>(x, y, cost);
    }

    template<int cost_from>
    inline void add_cell(const Block<T>& b, const Cell<T>* prev_c = nullptr, const Cell<T>* next_c = nullptr)
    {
      add_cell<diagonal, cost_from>(b, prev_c, next_c, b.right, b.top);
    }
    template<int cost_from>
    inline void add_cell(const Block<T>& b, const Diagonal<T>* prev_d, const Diagonal<T>* next_d)
    {
      if((cost_from & horizontal) == 0)
        add_cell<diagonal, cost_from>(b, nullptr, next_d->last_added_cell, b.right, b.top);
      else if((cost_from & vertical) == 0)
        add_cell<diagonal, cost_from>(b, prev_d->last_added_cell, nullptr, b.right, b.top);
      else
        add_cell<diagonal, cost_from>(b, prev_d->last_added_cell, next_d->last_added_cell, b.right, b.top);
    }
    template<int cost_from>
    inline void add_cell(const Block<T>& b, const Diagonal<T>& prev_d, const Diagonal<T>& next_d)
    {
      add_cell<cost_from>(b, prev_d.last_added_cell, next_d.last_added_cell);
    }

    template<int target>
    inline T cost_to(const int_fast32_t x, const int_fast32_t y, const T block_cost) const
    {
      const Cell<T>& last = last_relevant();
      assert(!(target & horizontal) || (last.y == y));
      assert(!(target & vertical) || (last.x == x));
      assert(!(target & diagonal) || (last.y - y == last.x - x));
      if(target & horizontal){
        return last.horizontal_cost(x, block_cost);
      } else {
        return last.vertical_cost(y, block_cost);
      }
    }

    // return top-row/right-col of the intersection between the diagonal and a block
    inline int_fast32_t vert_intersect(const int_fast32_t y) const { return y - x_offset; }
    inline int_fast32_t hori_intersect(const int_fast32_t x) const { return x + x_offset; }

    inline int_fast32_t top_intersect(const RawBlock& b) const { return hori_intersect(b.top); }
    inline int_fast32_t right_intersect(const RawBlock& b) const { return vert_intersect(b.right); }
    inline int_fast32_t bottom_intersect(const RawBlock& b) const { return hori_intersect(b.bottom); }
    inline int_fast32_t left_intersect(const RawBlock& b) const { return vert_intersect(b.left); }

    inline const Cell<T>& last_relevant() const
    {
      assert(last_added_cell != nullptr);
      DEBUG(std::cout << "last relevant cell of diagonal "<<x_offset<<": "<<*last_added_cell<<"\n");
      return *last_added_cell;
    }

    friend std::ostream& operator<<(std::ostream& os, const Diagonal& d) {  return os<<"["<<d.x_offset<<" last: "<<*(d.last_added_cell)<<"]\n"; }
  };

  template<class T = double>
  using DiagonalIter = typename std::list<Diagonal<T>>::iterator;

  template<class T = double>
  using DiagList = std::list<Diagonal<T>>;

}

