
#pragma once

#include "utils.hpp"
#include "block.hpp"
#include "diagonal.hpp"
#include "trace.hpp"

namespace kappa {

  template<class T = double>
  class Solver
  {
    const uint_fast32_t runs_x, runs_y;
    uint_fast32_t index;
    
    BlockVec<T> blocks;
    DiagList<T> diagonals;
    DiagonalIter<T> d;


    void treat_first_block(DiagonalIter<T>& d, const uint_fast32_t i, const uint_fast32_t j)
    {
      const Block<T>& b = blocks[index++];
      assert(j != 0);
      assert(i == 0);
      // treat first block of non-first row
      if(*d < b){
        const DiagonalIter<T> next_d = std::next(d);
        const int_fast32_t x_diff = b.x_offset();
        // if first first diagonal d comes before b, then this is the (0,0)-diagonal
        const DiagonalIter<T> new_d = diagonals.emplace(next_d, x_diff, runs_x, runs_y);

        DEBUG(std::cout << "adding new diagonal "<<new_d->x_offset<<" right of main diagonal\n");
        d->template add_cell<horizontal, diagonal>(b, nullptr, nullptr, b.top, b.top);
        Tracer<T>(runs_x, blocks, new_d).trace_new_diagonal(*d, *next_d, 0, j);
        d = next_d;
      } else if(*d == b){
        DEBUG(std::cout << "adding new cell to main diagonal "<<d->x_offset<<" at ("<<b.right<<","<<b.top<<")\n");
        d->template add_cell<diagonal>(b);
        ++d;
      } else { // *d > b
        DEBUG(std::cout<<"adding simple diagonal for offset "<<b.x_offset()<<" (block cost "<<b.cost<<")\n");
        if(b.right >= b.bottom) d->template add_cell<vertical, diagonal>(b, nullptr, nullptr, b.right, b.right);
        diagonals.emplace(d, b.x_offset(), runs_x, runs_y)->template add_cell<vertical>(b, nullptr, d->last_added_cell);
      }
    }

    void treat_other_block(DiagonalIter<T>& d, const uint_fast32_t i, const uint_fast32_t j)
    {
      const Block<T>& b = blocks[index++];
      assert(j != 0);
      assert(i != 0);
      assert(d != diagonals.begin());

      // step 1: go left -> right on the top row, adding intersections of our top row with the diagonals, starting at d
      const Cell<T>* last_added = std::prev(d)->last_added_cell;
      while(*d < b){
        d->template add_cell<horizontal, horizontal | diagonal>(b, last_added, nullptr, d->top_intersect(b), b.top);
        last_added = d->last_added_cell;
        ++d;
        assert(d != diagonals.end());
      }
      // step 2: go top->bottom and then bottom->top, adding intersections of our right column with the diagonals
      DEBUG(std::cout << "done treating top row, now at diagonal "<<d->x_offset<<'\n');
      if(d->bottom_intersect(b) <= b.right){
        // initial == end  iff  we are a block of the bottom row (and our upper-right cell's cost is defined by a horizontal path)
        DiagonalIter<T> vert_d = d;
        // if there is a block below us, then that block has a diagonal through its upper-right corner, so we reach that before reaching end()
        // step 2a: go to the bottom of the right column
        while(vert_d->right_intersect(b) >= b.bottom) ++vert_d;
        assert(vert_d != diagonals.end());
        // step 2b: go up to initial again, adding the intersection cells
        while(vert_d != d){
          const DiagonalIter<T> previous = vert_d;
          --vert_d;
          if(*vert_d == b) break;
          vert_d->template add_cell<vertical, vertical | diagonal>(b, nullptr, previous->last_added_cell, b.right, vert_d->right_intersect(b));
        }
      }
      // add diagonal of this block before d if necessary
      if(*d != b){
        const DiagonalIter<T> new_d = diagonals.emplace(d, b.x_offset(), runs_x, runs_y);
        Tracer<T>(runs_x, blocks, new_d).trace_new_diagonal(*std::prev(new_d), *d, i, j);
      } else {
        // if a diagonal with this offset has already been created, we know that there are diagonals before and after d
        d->template add_cell<vertical | horizontal | diagonal>(b, *std::prev(d), *std::next(d));
        ++d;
      }
    }

    void treat_first_row(DiagonalIter<T>& d, const uint_fast32_t j)
    {
      assert(j == 0);

      const Cell<T> initial_cell = {-1, -1, 0};
      d = diagonals.emplace(diagonals.end(), 0, runs_x, runs_y);
      // this is pure cheating, since the cell is not owned by the diagonal, but we can get away with it :)
      d->last_added_cell = &initial_cell;
     
      // init lowest row of blocks as long as they intersect the (0,0)-diagonal
      do{
        DEBUG(std::cout << "treating block["<<index<<"] on first row\n");
        const Block<T>& b = blocks[index++];
        if(b.right <= b.top){
          if(b.right != b.top){
            d->template add_cell<vertical, diagonal>(b, nullptr, nullptr, b.right, b.right);
            // emplace the new diagonal before the main diagonal
            const DiagonalIter<T> new_d = diagonals.emplace(d, b.x_offset(), runs_x, runs_y);
            if(index == 1)
              new_d->template add_cell<vertical>(b, nullptr, d->last_added_cell);
            else Tracer<T>(runs_x, blocks, new_d).template trace_new_diagonal<false, true>(nullptr, &(*d), index - 1, 0);
          } else {
            d->template add_cell<diagonal>(b);
            ++d;
            break;
          }
        } else { // b.right > b.top
          d->template add_cell<horizontal, diagonal>(b, nullptr, nullptr, b.top, b.top);
          // cost for the upper-right corner comes from the (-1,-1)-diagonal
          const DiagonalIter<T> new_d = diagonals.emplace(diagonals.end(), b.x_offset(), runs_x, runs_y);
          if(index == 1)
            new_d->template add_cell<horizontal>(b, &(*d), nullptr);
          else Tracer<T>(runs_x, blocks, new_d).template trace_new_diagonal<true, false>(&(*d), nullptr, index - 1, 0);
          d = diagonals.end();
          break;
        }
      } while(index != runs_x);

      if(index < runs_x){
        // finish the lowest row
        while(index < runs_x){
          const Block<T>& b = blocks[index++];
          const Cell<T>* last_added = diagonals.back().last_added_cell;
          diagonals.emplace_back(b.x_offset(), runs_x, runs_y).template add_cell<horizontal>(b, last_added, nullptr);
        }
        d = diagonals.end();
      }
    }

    void treat_other_row(DiagonalIter<T>& d, const uint_fast32_t j)
    {
      assert(j != 0);
      treat_first_block(d, 0, j);
      // treat the rest of the blocks
      for(uint_fast32_t i = 1; i != runs_x; ++i) treat_other_block(d, i, j);
    }

  public:

    T compute_dtw()
    {   
      index = 0;

      treat_first_row(d, 0);
      // then init all others
      for(uint_fast32_t j = 1; j != runs_y; ++j) {
        d = diagonals.begin();
        treat_other_row(d, j);
      }
      // the final cost is stored in the upper-right corner of the last block
      return (--d)->last_relevant().cost;
    }

    Solver(const RLEncoding<T>& input_x, const RLEncoding<T>& input_y):
      runs_x(input_x.size()), runs_y(input_y.size())
    {
      assert((runs_x > 0) && (runs_y > 0));
      // setup blocks
      blocks.reserve(runs_x * runs_y);
      int_fast32_t top = -1, right = -1;
      int_fast32_t bottom, left;
      for(uint_fast32_t j = 0; j < runs_y; ++j){
        const RunlengthAndValue<T>& y = input_y[j];
        bottom = top + 1;
        top += y.first;
        right = -1;
        for(uint_fast32_t i = 0; i < runs_x; ++i){
          const RunlengthAndValue<T>& x = input_x[i];
          const T diff = x.second - y.second;
          left = right + 1;
          right += x.first;
          blocks.emplace_back(bottom, left, top, right, diff * diff);
        }
      }
    }

#ifdef STATISTICS
  uint_fast32_t get_kappa() const
  {
    assert(!diagonals.empty() && "to call get_kappa(), make sure that compute_dtw() has run before");
    uint_fast32_t k = 0;
    for(const auto& d: diagonals) k += d.kappa;
    return k;
  }
#endif

  };

}

