
#include <iostream>
#include <fstream>
#include "dtw.hpp"

#include "kn.hpp"
#include "nn.hpp"

#include<cstdlib>
#include<ctime>
#include<algorithm>

#include "ucrParse.hpp"

using CostType = double;

template<class T = double>
RLEncoding<T> generate(const size_t len, const uint_fast32_t block_size)
{
  RLEncoding<T> x;
  for(size_t i = 0; i != len; ++i)
    x.emplace_back(std::max((uint_fast32_t)0u, block_size + rand() % 5 - 2), rand() % 10);
  return x;
}

template<class T = double>
void run_test(const RLEncoding<T>& x, const RLEncoding<T>& y)
{
  DEBUG(std::cout << "DTW of\n"<<x<<"\n"<<y<<"\n");
  clock_t start_clock = std::clock();
  kappa::Solver<CostType> kappa_solver(x,y);
  const T d1 = kappa_solver.compute_dtw();
  double time_spent = 1000.0 * (std::clock() - start_clock) / CLOCKS_PER_SEC;

  std::cout << "O(kappa)-algo: cost: "<< d1 << " time: "<<time_spent<<"ms";
#ifdef STATISTICS
  std::cout << " (kappa: "<< kappa_solver.get_kappa()<<')';
#endif
  std::cout << std::endl;

  start_clock = std::clock();
  const T d2 = kn::compute_dtw(x, y);
  time_spent = 1000.0 * (std::clock() - start_clock) / CLOCKS_PER_SEC;
  std::cout << "O(kn)-algo:    cost: "<< d2 << " time: "<<time_spent<<"ms\n";

  if((d1 != d2)){
    std::cout << "conflicting costs: "<<d1<<" & "<<d2<<"\nTHIS IS A BUG, PLEASE REPORT\n";
    exit(1);
  }

  start_clock = std::clock();
  const T d3 = nn::compute_dtw(x, y);
  time_spent = 1000.0 * (std::clock() - start_clock) / CLOCKS_PER_SEC;
  std::cout << "O(n^2)-algo:   cost: "<< d3 << " time: "<<time_spent<<"ms\n";

  if((d1 != d3)){
    std::cout << "conflicting costs: "<<d1<<" & "<<d3<<"\nTHIS IS A BUG, PLEASE REPORT\n";
    exit(1);
  }
}

template<class T = double>
void run_tests(const ssize_t num_iterations, const uint_fast32_t k, const uint_fast32_t block_size)
{
  for(int i = 0; i != num_iterations; ++i){
    RLEncoding<T> x, y;
    run_test(generate(k, block_size), generate(k, block_size));
    std::cout << "\n\n";
  }
}

int main(int argc, char** argv){ 
  //std::srand(std::time(nullptr));
  if(argc > 3){
    run_tests(atoi(argv[3]), atoi(argv[1]), atoi(argv[2]));
  } else if((argc > 1) && (std::string("-h") != argv[1]) && (std::string("--help") != argv[1])){
    std::cout << "trying to load 2 time series from file "<<argv[1]<<"\n";
    std::ifstream _in(argv[1]);
    if(_in.is_open()) {
      const RLEncoding<CostType> x = time_seq_to_RLE(akt::parseUCR<CostType>(_in, 1).back());
      const RLEncoding<CostType> y = time_seq_to_RLE(akt::parseUCR<CostType>(_in, 1).back());
      std::cout << "RLEncodings of length "<<x.size()<<" & "<<y.size()<<'\n';
      run_test(x, y);
      return 0;
    } else {
      std::cerr << "failed to open file " << argv[1] << " for reading\n";
      return 1;
    }
  } else {
    std::cerr << "usage: "<<argv[0]<<" <filename>       - compute dtw of first two sequences in filename (in UCR format)\n";
    std::cerr << "       "<<argv[0]<<" <k> <rl> <iter>  - speed test: iter iterations of k runs against k runs, each run of length rl+/-2\n";
    std::cerr << "       "<<argv[0]<<" <-h|--help>      - show this help screen\n";
    return 1;
  }
}
