
#pragma once

#include "diagonal.hpp"

namespace kappa {

  template<class T = double>
  struct CountingRevCellIter: public std::pair<typename Diagonal<T>::RevCellIter, ssize_t>
  {
    using Parent = std::pair<typename Diagonal<T>::RevCellIter, ssize_t>;
    using Parent::Parent;
    using Parent::first;
    using Parent::second;

    CountingRevCellIter(const typename Diagonal<T>::CellStorage& lst): Parent(std::next(lst.rbegin()), lst.size() - 1) {}
    CountingRevCellIter& operator++() { ++first; --second; return *this; }
    const Cell<T>& operator*() const { return *first; }
    operator const Cell<T>*() const { return &(*first); }
    CountingRevCellIter next() const { CountingRevCellIter result(*this); return ++result; }
  };


  template<class T = double>
  struct Tracer
  {
    const uint_fast32_t runs_x;
    const BlockVec<T>& blocks;
    const DiagonalIter<T>& d;

    Tracer(const uint_fast32_t _runs_x, const BlockVec<T>& _blocks, const DiagonalIter<T>& _d):
      runs_x(_runs_x), blocks(_blocks), d(_d) {}

    inline uint_fast32_t idx(const uint_fast32_t x, uint_fast32_t y) const { return runs_x * y + x; }


    // after construction, setup the relevant cells, starting at block[i,j]
    template<bool has_prev, bool has_next, bool above_main_diagonal, direction from>
    void trace_cells(const CountingRevCellIter<T>* const prev_hori,
                     const CountingRevCellIter<T>* const next_vert,
                     const int_fast32_t i,
                     const int_fast32_t j) const
    {
      assert((i >= 0) && (j >= 0));
      assert(has_prev ? prev_hori->second >= 0 : true);
      assert(has_next ? next_vert->second >= 0 : true);

      // make recursive call, depending on where we exit b
      const Block<T>& b = blocks[idx(i,j)];
      const int_fast32_t bottom_intersect = d->bottom_intersect(b);
      
      // if i,j > 0 and bottom_intersect == b.left, then this diagonal should have been added before
      assert((bottom_intersect != b.left) || (i == 0) || (j == 0));

      // recurse only if we haven't yet hit the bottom row (if we're below the main diagonal) or first column (if we're above the main diagonal)
      if( (above_main_diagonal && (i != 0)) || (!above_main_diagonal && (j != 0)) ){
        CountingRevCellIter<T> _prev_hori, _next_vert;
        if(has_prev) _prev_hori = *prev_hori;
        if(has_next) _next_vert = *next_vert;
        if(has_prev && (from & vertical)) ++_prev_hori;
        if(has_next && (from & horizontal)) ++_next_vert;

        if(bottom_intersect < b.left){
          trace_cells<has_prev, has_next, above_main_diagonal, horizontal>(&_prev_hori, &_next_vert, i - 1, j);
        } else {
          trace_cells<has_prev, has_next, above_main_diagonal, vertical>(&_prev_hori, &_next_vert, i, j - 1);
        }
      }

      // add the relevant cell depending on where we entered the block b
      if(from & vertical){
        // if we came from the block above (vertical), then we add a new horizontal cell (top row of b)
        // the followingn check is unnecessary due to the layout of the DP:
          d->template add_cell<horizontal, horizontal | diagonal>(b, *prev_hori, nullptr, d->top_intersect(b), b.top);
      } else {
        // if we came from the block to the right (horizontal), then we add a new vertical cell (right column of b)
        if(has_next ? next_vert->second != 0 : false){
          d->template add_cell<vertical, vertical | diagonal>(b, nullptr, *next_vert, b.right, d->right_intersect(b));
        } else
          if(d->last_added_cell != nullptr)
            d->template add_cell<vertical, diagonal>(b, nullptr, nullptr, b.right, d->right_intersect(b));
      }
    }

    template<bool has_prev, bool has_next>
    inline void trace_cells(const CountingRevCellIter<T>* prev_hori,
                            const CountingRevCellIter<T>* next_vert,
                            const int_fast32_t i,
                            const int_fast32_t j)
    {
      const Block<T>& b = blocks[idx(i,j)];
      if((d->bottom_intersect(b) < b.left) && (i != 0)){// new_d intersects left border of b
        if(d->x_offset <= 0)
          trace_cells<has_prev, has_next, true, horizontal>(prev_hori, next_vert, i - 1, j);
        else
          trace_cells<has_prev, has_next, false, horizontal>(prev_hori, next_vert, i - 1, j);
      } else if(j != 0){ // new_d intersects bottom border of b
        if(d->x_offset <= 0)
          trace_cells<has_prev, has_next, true, vertical>(prev_hori, next_vert, i, j - 1);
        else
          trace_cells<has_prev, has_next, false, vertical>(prev_hori, next_vert, i, j - 1);
      }
    }

    // setup the relevant cells on the diagonal d
    template<bool has_pred, bool has_next>
    void trace_new_diagonal(const Diagonal<T>* prev_d,
                            const Diagonal<T>* next_d,
                            const int_fast32_t i,
                            const int_fast32_t j)
    {
      assert(has_pred || has_next);

      const Block<T>& b = blocks[idx(i,j)];
      if(has_pred && has_next){
        const CountingRevCellIter<T> prev_hori(prev_d->relevant_cells_hori);
        const CountingRevCellIter<T> next_vert(next_d->relevant_cells_vert);
        trace_cells<true,true>(&prev_hori, &next_vert, i, j);
        d->template add_cell<horizontal | vertical | diagonal>(b, prev_d, next_d);
      } else if(has_pred){ // !has_next
        const CountingRevCellIter<T> prev_hori(prev_d->relevant_cells_hori);
        trace_cells<true,false>(&prev_hori, nullptr, i, j);
        d->template add_cell<horizontal | diagonal>(b, prev_d, nullptr);
      } else { // has_next
        const CountingRevCellIter<T> next_vert(next_d->relevant_cells_vert);
        trace_cells<false,true>(nullptr, &next_vert, i, j);
        d->template add_cell<vertical | diagonal>(b, nullptr, next_d);
      }
    }

    inline void trace_new_diagonal(const Diagonal<T>& prev_d,
                                   const Diagonal<T>& next_d,
                                   const int_fast32_t i,
                                   const int_fast32_t j)
    {
      return trace_new_diagonal<true,true>(&prev_d, &next_d, i, j);
    }

  };
}

